from unittest.mock import Mock, patch

import pytest
from requests.exceptions import Timeout

from ...conf.test import override_dynamic_settings
from .. import exceptions
from ..client import REQUESTS_TIMEOUT, get_user_data

ACCESS_TOKEN = "acc3ss-t0k3n"


@pytest.fixture
def mock_request(dynamic_settings):
    return Mock(settings=dynamic_settings)


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="GET",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_get_request_with_token_in_query_string(
    mock_request,
):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/user?atoken={ACCESS_TOKEN}",
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="GET",
    oauth2_user_token_name="Authentication",
    oauth2_user_token_location="HEADER",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_get_request_with_token_in_header(mock_request):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/user",
            headers={"Authentication": ACCESS_TOKEN},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="GET",
    oauth2_user_token_name="Authentication",
    oauth2_user_token_location="HEADER_BEARER",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_get_request_with_bearer_token_in_header(
    mock_request,
):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/user",
            headers={"Authentication": f"Bearer {ACCESS_TOKEN}"},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="POST",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_post_request_with_token_in_query_string(
    mock_request,
):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        post_mock.assert_called_once_with(
            f"https://example.com/oauth2/user?atoken={ACCESS_TOKEN}",
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="POST",
    oauth2_user_token_name="Authentication",
    oauth2_user_token_location="HEADER",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_post_request_with_token_in_header(mock_request):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        post_mock.assert_called_once_with(
            f"https://example.com/oauth2/user",
            headers={"Authentication": ACCESS_TOKEN},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="POST",
    oauth2_user_token_name="Authentication",
    oauth2_user_token_location="HEADER_BEARER",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_post_request_with_bearer_token_in_header(
    mock_request,
):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        post_mock.assert_called_once_with(
            f"https://example.com/oauth2/user",
            headers={"Authentication": f"Bearer {ACCESS_TOKEN}"},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/user",
    oauth2_user_method="POST",
    oauth2_user_token_name="Authentication",
    oauth2_user_token_location="HEADER_BEARER",
    oauth2_user_extra_headers="Accept: application/json\nApi-Ver:1234",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_is_returned_using_post_request_with_extra_headers(
    mock_request,
):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        post_mock.assert_called_once_with(
            f"https://example.com/oauth2/user",
            headers={
                "Authentication": f"Bearer {ACCESS_TOKEN}",
                "Accept": "application/json",
                "Api-Ver": "1234",
            },
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data?type=user",
    oauth2_user_method="GET",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
    oauth2_json_id_path="id",
    oauth2_json_name_path="name",
    oauth2_json_email_path="email",
    oauth2_json_avatar_path="avatar",
)
def test_user_data_request_with_token_in_url_respects_existing_querystring(
    mock_request,
):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=user_data,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, user_data)

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/data?type=user&atoken={ACCESS_TOKEN}",
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data?type=user",
    oauth2_user_method="GET",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
    oauth2_json_id_path="id",
    oauth2_json_name_path="user.profile.name",
    oauth2_json_email_path="user.profile.email",
    oauth2_json_avatar_path="user.profile.avatar",
)
def test_user_data_json_values_are_mapped_to_result(mock_request):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    json_response = {
        "id": user_data["id"],
        "user": {
            "profile": {
                "name": user_data["name"],
                "email": user_data["email"],
                "avatar": user_data["avatar"],
            }
        },
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=json_response,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, json_response)

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/data?type=user&atoken={ACCESS_TOKEN}",
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data?type=user",
    oauth2_user_method="GET",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
    oauth2_json_id_path="id",
    oauth2_json_name_path="user.profile.name",
    oauth2_json_email_path="user.profile.email",
    oauth2_json_avatar_path="user.profile.avatar",
)
def test_user_data_json_values_are_none_when_not_found(mock_request):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": "https://example.com/avatar.png",
    }

    json_response = {
        "profile_id": user_data["id"],
        "user": {
            "data": {
                "name": user_data["name"],
                "email": user_data["email"],
                "avatar": user_data["avatar"],
            }
        },
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=json_response,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (
            {
                "id": None,
                "name": None,
                "email": None,
                "avatar": None,
            },
            json_response,
        )

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/data?type=user&atoken={ACCESS_TOKEN}",
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data?type=user",
    oauth2_user_method="GET",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
    oauth2_json_id_path="id",
    oauth2_json_name_path="user.profile.name",
    oauth2_json_email_path="user.profile.email",
    oauth2_json_avatar_path="",
)
def test_user_data_skips_avatar_if_path_is_not_set(mock_request):
    user_data = {
        "id": "7dds8a7dd89sa",
        "name": "Aerith",
        "email": "aerith@example.com",
        "avatar": None,
    }

    json_response = {
        "id": user_data["id"],
        "user": {
            "profile": {
                "name": user_data["name"],
                "email": user_data["email"],
                "avatar": "https://example.com/avatar.png",
            }
        },
    }

    get_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=json_response,
            ),
        ),
    )

    with patch("requests.get", get_mock):
        assert get_user_data(mock_request, ACCESS_TOKEN) == (user_data, json_response)

        get_mock.assert_called_once_with(
            f"https://example.com/oauth2/data?type=user&atoken={ACCESS_TOKEN}",
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data",
    oauth2_user_method="POST",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
)
def test_exception_is_raised_if_user_data_request_times_out(mock_request):
    post_mock = Mock(side_effect=Timeout())

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2UserDataRequestError):
            get_user_data(mock_request, ACCESS_TOKEN)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data",
    oauth2_user_method="POST",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
)
def test_exception_is_raised_if_user_data_request_response_is_not_200(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=400,
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2UserDataResponseError):
            get_user_data(mock_request, ACCESS_TOKEN)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data",
    oauth2_user_method="POST",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
)
def test_exception_is_raised_if_user_data_request_response_is_not_json(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                side_effect=ValueError(),
            ),
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2UserDataJSONError):
            get_user_data(mock_request, ACCESS_TOKEN)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_user_url="https://example.com/oauth2/data",
    oauth2_user_method="POST",
    oauth2_user_token_name="atoken",
    oauth2_user_token_location="QUERY",
)
def test_exception_is_raised_if_user_data_request_response_json_is_not_object(
    mock_request,
):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=["json", "list"],
            ),
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2UserDataJSONError):
            get_user_data(mock_request, ACCESS_TOKEN)

        post_mock.assert_called_once()



from django.utils.translation import pgettext_lazy


class MisagoAdminExtension:
    def register_navigation_nodes(self, site):
        site.add_node(name=pgettext_lazy("admin node", "Dashboard"), icon="fa fa-home")



import re

from django.forms import ValidationError
from django.utils.translation import pgettext, pgettext_lazy

from . import basefields


class BioField(basefields.UrlifiedTextareaProfileField):
    fieldname = "bio"
    label = pgettext_lazy("bio profile field", "Bio")


class RealNameField(basefields.TextProfileField):
    fieldname = "real_name"
    label = pgettext_lazy("real_name profile field", "Real name")


class LocationField(basefields.TextProfileField):
    fieldname = "location"
    label = pgettext_lazy("location profile field", "Location")


class GenderField(basefields.ChoiceProfileField):
    fieldname = "gender"
    label = pgettext_lazy("gender profile field", "Gender")

    choices = (
        ("", pgettext_lazy("gender profile field choice", "Not specified")),
        ("secret", pgettext_lazy("gender profile field choice", "Not telling")),
        ("female", pgettext_lazy("gender profile field choice", "Female")),
        ("male", pgettext_lazy("gender profile field choice", "Male")),
        ("enby", pgettext_lazy("gender profile field choice", "Non-binary")),
    )


class WebsiteField(basefields.UrlProfileField):
    fieldname = "website"
    label = pgettext_lazy("website profile field", "Website")
    help_text = pgettext_lazy(
        "website profile field",
        'If you own a website you wish to share on your profile, you may enter its address here. Remember that, for it to be valid, it should start with either "http://" or "https://".',
    )


class SkypeIdField(basefields.TextProfileField):
    fieldname = "skype"
    label = pgettext_lazy("skype id profile field", "Skype ID")


class TwitterHandleField(basefields.TextProfileField):
    fieldname = "twitter"
    label = pgettext_lazy("twitter handle profile field", "X (formerly Twitter) handle")

    def get_help_text(self, user):
        return pgettext_lazy(
            "twitter handle profile field",
            'If you own an X account, here you may enter your X handle for other users to find you. Starting your handle with "@" sign is optional. Either "@%(slug)s" or "%(slug)s" are valid values.',
        ) % {"slug": user.slug}

    def get_value_display_data(self, request, user, value):
        return {"text": "@%s" % value, "url": "https://twitter.com/%s" % value}

    def clean(self, request, user, data):
        data = data.lstrip("@")
        if data and not re.search("^[A-Za-z0-9_]+$", data):
            raise ValidationError(
                pgettext(
                    "twitter handle profile field",
                    "This is not a valid X handle.",
                )
            )
        return data


class JoinIpField(basefields.TextProfileField):
    fieldname = "join_ip"
    label = pgettext_lazy("join ip profile field", "Join IP")
    readonly = True

    def get_value_display_data(self, request, user, value):
        if not request.user_acl.get("can_see_users_ips"):
            return None

        if not user.joined_from_ip:
            return None

        return {"text": user.joined_from_ip}



from django.contrib.auth import get_user_model
from django.db import IntegrityError, transaction
from django.utils import timezone
from django.utils.translation import pgettext_lazy

from ..users.setupnewuser import setup_new_user
from .exceptions import OAuth2UserDataValidationError, OAuth2UserIdNotProvidedError
from .models import Subject
from .validation import validate_user_data

User = get_user_model()


def get_user_from_data(request, user_data, user_data_raw):
    if not user_data["id"]:
        raise OAuth2UserIdNotProvidedError()

    user = get_user_by_subject(user_data["id"])
    if not user and user_data["email"]:
        user = get_user_by_email(user_data["id"], user_data["email"])

    created = not bool(user)

    cleaned_data = validate_user_data(request, user, user_data, user_data_raw)

    try:
        with transaction.atomic():
            if not user:
                user = create_new_user(request, cleaned_data)
            else:
                update_existing_user(user, cleaned_data)
    except IntegrityError as error:
        raise_validation_error_from_integrity_error(error)

    return user, created


def get_user_by_subject(user_id):
    try:
        subject = Subject.objects.select_related("user", "user__ban_cache").get(
            sub=user_id
        )
        subject.last_used_on = timezone.now()
        subject.save(update_fields=["last_used_on"])
        return subject.user
    except Subject.DoesNotExist:
        return None


def get_user_by_email(user_id, user_email):
    try:
        user = User.objects.get_by_email(user_email)
        Subject.objects.create(sub=user_id, user=user)
    except User.DoesNotExist:
        return None


def create_new_user(request, user_data):
    activation_kwargs = {}
    if request.settings.account_activation == "admin":
        activation_kwargs = {"requires_activation": User.ACTIVATION_ADMIN}

    user = User.objects.create_user(
        user_data["name"],
        user_data["email"],
        joined_from_ip=request.user_ip,
        **activation_kwargs,
    )

    setup_new_user(request.settings, user, avatar_url=user_data["avatar"])
    Subject.objects.create(sub=user_data["id"], user=user)

    return user


def update_existing_user(user, user_data):
    save_changes = False

    if user.username != user_data["name"]:
        user.set_username(user_data["name"])
        save_changes = True

    if user.email != user_data["email"]:
        user.set_email(user_data["email"])
        save_changes = True

    if save_changes:
        user.save()


def raise_validation_error_from_integrity_error(error):
    error_str = str(error)

    if "misago_users_user_email_hash_key" in error_str:
        raise OAuth2UserDataValidationError(
            error_list=[
                pgettext_lazy(
                    "oauth2 error",
                    "Your e-mail address returned by the provider is not available for use on this site.",
                )
            ]
        )

    if "misago_users_user_slug_key" in error_str:
        raise OAuth2UserDataValidationError(
            error_list=[
                pgettext_lazy(
                    "oauth2 error",
                    "Your username returned by the provider is not available for use on this site.",
                )
            ]
        )



from django.test import TestCase
from django.utils import timezone

from ...categories.models import Category
from ...users.test import create_test_user
from ..models import Post, Thread, ThreadParticipant
from ..participants import (
    add_participants,
    has_participants,
    make_participants_aware,
    set_owner,
    set_users_unread_private_threads_sync,
)


class ParticipantsTests(TestCase):
    def setUp(self):
        datetime = timezone.now()

        self.category = Category.objects.all_categories()[:1][0]
        self.thread = Thread(
            category=self.category,
            started_on=datetime,
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=datetime,
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        self.thread.set_title("Test thread")
        self.thread.save()

        post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster_name="Tester",
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        self.thread.first_post = post
        self.thread.last_post = post
        self.thread.save()

    def test_has_participants(self):
        """has_participants returns true if thread has participants"""
        users = [
            create_test_user("User", "user@example.com"),
            create_test_user("Other_User", "otheruser@example.com"),
        ]

        self.assertFalse(has_participants(self.thread))

        ThreadParticipant.objects.add_participants(self.thread, users)
        self.assertTrue(has_participants(self.thread))

        self.thread.threadparticipant_set.all().delete()
        self.assertFalse(has_participants(self.thread))

    def test_make_threads_participants_aware(self):
        """
        make_participants_aware sets participants_list and participant
        annotations on list of threads
        """
        user = create_test_user("User", "user@example.com")
        other_user = create_test_user("Other_User", "otheruser@example.com")

        self.assertFalse(hasattr(self.thread, "participants_list"))
        self.assertFalse(hasattr(self.thread, "participant"))

        make_participants_aware(user, [self.thread])

        self.assertFalse(hasattr(self.thread, "participants_list"))
        self.assertTrue(hasattr(self.thread, "participant"))
        self.assertIsNone(self.thread.participant)

        ThreadParticipant.objects.set_owner(self.thread, user)
        ThreadParticipant.objects.add_participants(self.thread, [other_user])

        make_participants_aware(user, [self.thread])

        self.assertFalse(hasattr(self.thread, "participants_list"))
        self.assertEqual(self.thread.participant.user, user)

    def test_make_thread_participants_aware(self):
        """
        make_participants_aware sets participants_list and participant
        annotations on thread model
        """
        user = create_test_user("User", "user@example.com")
        other_user = create_test_user("Other_User", "otheruser@example.com")

        self.assertFalse(hasattr(self.thread, "participants_list"))
        self.assertFalse(hasattr(self.thread, "participant"))

        make_participants_aware(user, self.thread)

        self.assertTrue(hasattr(self.thread, "participants_list"))
        self.assertTrue(hasattr(self.thread, "participant"))

        self.assertEqual(self.thread.participants_list, [])
        self.assertIsNone(self.thread.participant)

        ThreadParticipant.objects.set_owner(self.thread, user)
        ThreadParticipant.objects.add_participants(self.thread, [other_user])

        make_participants_aware(user, self.thread)

        self.assertEqual(self.thread.participant.user, user)
        for participant in self.thread.participants_list:
            if participant.user == user:
                break
        else:
            self.fail("thread.participants_list didn't contain user")

    def test_set_owner(self):
        """set_owner sets user as thread owner"""
        user = create_test_user("User", "user@example.com")

        set_owner(self.thread, user)

        owner = self.thread.threadparticipant_set.get(is_owner=True)
        self.assertEqual(user, owner.user)

    def test_set_users_unread_private_threads_sync(self):
        """
        set_users_unread_private_threads_sync sets sync_unread_private_threads
        flag on users provided to true
        """
        users = [
            create_test_user("User", "user@example.com"),
            create_test_user("Other_User", "otheruser@example.com"),
        ]

        set_users_unread_private_threads_sync(users=users)
        for user in users:
            user.refresh_from_db()
            assert user.sync_unread_private_threads

    def test_set_participants_unread_private_threads_sync(self):
        """
        set_users_unread_private_threads_sync sets sync_unread_private_threads
        flag on participants provided to true
        """
        users = [
            create_test_user("User", "user@example.com"),
            create_test_user("Other_User", "otheruser@example.com"),
        ]

        participants = [ThreadParticipant(user=u) for u in users]

        set_users_unread_private_threads_sync(participants=participants)
        for user in users:
            user.refresh_from_db()
            assert user.sync_unread_private_threads

    def test_set_participants_users_unread_private_threads_sync(self):
        """
        set_users_unread_private_threads_sync sets sync_unread_private_threads
        flag on users and participants provided to true
        """
        users = [create_test_user("User", "user@example.com")]
        participants = [ThreadParticipant(user=u) for u in users]
        users.append(create_test_user("Other_User", "otheruser@example.com"))

        set_users_unread_private_threads_sync(users=users, participants=participants)
        for user in users:
            user.refresh_from_db()
            assert user.sync_unread_private_threads

    def test_set_users_unread_private_threads_sync_exclude_user(self):
        """exclude_user kwarg works"""
        users = [
            create_test_user("User", "user@example.com"),
            create_test_user("Other_User", "otheruser@example.com"),
        ]

        set_users_unread_private_threads_sync(users=users, exclude_user=users[0])

        [i.refresh_from_db() for i in users]
        assert users[0].sync_unread_private_threads is False
        assert users[1].sync_unread_private_threads

    def test_set_users_unread_private_threads_sync_noop(self):
        """excluding only user is noop"""
        user = create_test_user("User", "user@example.com")

        with self.assertNumQueries(0):
            set_users_unread_private_threads_sync(users=[user], exclude_user=user)

        user.refresh_from_db()
        assert user.sync_unread_private_threads is False


def test_add_participants_triggers_notify_on_new_private_thread(
    mocker, user, other_user, private_thread
):
    notify_on_new_private_thread_mock = mocker.patch(
        "misago.threads.participants.notify_on_new_private_thread"
    )

    add_participants(user, private_thread, [user, other_user])

    notify_on_new_private_thread_mock.delay.assert_called_once_with(
        user.id, private_thread.id, [other_user.id]
    )



from typing import Optional, Protocol

from django.http import HttpRequest

from ...plugins.hooks import FilterHook
from ...users.models import User


class FilterUserDataHookAction(Protocol):
    """
    A standard Misago function used for filtering the user data, or the next
    filter function from another plugin.

    # Arguments

    ## `request: HttpRequest`

    The request object.

    ## `user: Optional[User]`

    A `User` object associated with `user_data["id"]` or `user_data["email"]`,
    or `None` if it's the user's first time signing in with OAuth and the user's
    account hasn't been created yet.

    ## `user_data: dict`

    A Python `dict` with user data extracted from the OAuth 2 server's response:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```

    # Return value

    A Python `dict` containing user data:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```
    """

    def __call__(
        self,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
    ) -> dict:
        pass


class FilterUserDataHookFilter(Protocol):
    """
    A function implemented by a plugin that can be registered in this hook.

    # Arguments

    ## `action: FilterUserDataHookAction`

    A standard Misago function used for filtering the user data, or the next
    filter function from another plugin.

    See the [action](#action) section for details.

    ## `request: HttpRequest`

    The request object.

    ## `user: Optional[User]`

    A `User` object associated with `user_data["id"]` or `user_data["email"]`,
    or `None` if it's the user's first time signing in with OAuth and the user's
    account hasn't been created yet.

    ## `user_data: dict`

    A Python `dict` with user data extracted from the OAuth 2 server's response:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```

    # Return value

    A Python `dict` containing user data:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```
    """

    def __call__(
        self,
        action: FilterUserDataHookAction,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
    ) -> dict:
        pass


class FilterUserDataHook(
    FilterHook[FilterUserDataHookAction, FilterUserDataHookFilter]
):
    """
    This hook wraps the standard function that Misago uses to filter a Python `dict`
    containing the user data extracted from the OAuth 2 server's response.

    User data filtering is part of the [user data validation by the OAuth 2
    client](./validate-user-data-hook.md), which itself is part of a function that
    creates a new user account or updates an existing one if user data has changed.

    Standard user data filtering doesn't validate the data but instead tries to
    improve it to increase its chances of passing the validation. It converts the
    `name` into a valid Misago username (e.g., `Łukasz Kowalski` becomes
    `Lukasz_Kowalski`). It also appends a random string at the end of the name if
    it's already taken by another user (e.g., `RickSanchez` becomes
    `RickSanchez_C137`). If the name is empty, a placeholder one is generated,
    e.g., `User_d6a9`. Lastly, it replaces an `email` with an empty string if
    it's `None`, to prevent a type error from being raised by e-mail validation
    that happens in the next step.

    Plugin filters can still raise Django's `ValidationError` on an invalid value
    instead of attempting to fix it if this is a preferable resolution.

    # Example

    The code below implements a custom filter function that extends the standard
    logic with additional user e-mail normalization for Gmail e-mails:

    ```python
    @filter_user_data_hook.append_filter
    def normalize_gmail_email(
        action, request: HttpRequest, user: Optional[User], user_data: dict
    ) -> dict:
        if (
            user_data["email"]
            and user_data["email"].lower().endswith("@gmail.com")
        ):
            # Dots in Gmail emails are ignored but frequently used by spammers
            new_user_email = user_data["email"][:-10].replace(".", "")
            user_data["email"] = new_user_email + "@gmail.com"

        # Call the next function in chain
        return action(user_data, request, user, user_data)
    ```
    """

    __slots__ = FilterHook.__slots__

    def __call__(
        self,
        action: FilterUserDataHookAction,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
    ) -> dict:
        return super().__call__(action, request, user, user_data)


filter_user_data_hook = FilterUserDataHook()



from unittest.mock import Mock

from ..context_processors import menus


def test_context_processor_adds_navbar_menu_to_context(
    navbar_menu_items, cache_versions
):
    result = menus(Mock(cache_versions=cache_versions))
    assert isinstance(result, dict)
    assert len(navbar_menu_items) == len(result["navbar_menu"])


def test_context_processor_adds_footer_menu_to_context(
    footer_menu_items, cache_versions
):
    result = menus(Mock(cache_versions=cache_versions))
    assert isinstance(result, dict)
    assert len(footer_menu_items) == len(result["footer_menu"])



from ..models import CacheVersion
from ..versions import invalidate_all_caches, invalidate_cache


def test_invalidating_cache_updates_cache_version_in_database(cache_version):
    invalidate_cache(cache_version.cache)
    updated_cache_version = CacheVersion.objects.get(cache=cache_version.cache)
    assert cache_version.version != updated_cache_version.version


def test_invalidating_all_caches_updates_cache_version_in_database(cache_version):
    invalidate_all_caches()
    updated_cache_version = CacheVersion.objects.get(cache=cache_version.cache)
    assert cache_version.version != updated_cache_version.version



import random
import sys

from django.core.management.base import BaseCommand
from faker import Factory

from ....core.management.progressbar import show_progress
from ....users.models import Ban
from ...bans import get_fake_username_ban, get_fake_email_ban, get_fake_ip_ban


class Command(BaseCommand):
    help = "Creates random fakey bans for testing purposes"

    def handle(self, *args, **options):
        try:
            fake_bans_to_create = int(args[0])
        except IndexError:
            fake_bans_to_create = 5
        except ValueError:
            self.stderr.write("\nOptional argument should be integer.")
            sys.exit(1)

        fake = Factory.create()
        ban_fakers = (get_fake_username_ban, get_fake_email_ban, get_fake_ip_ban)

        message = "Creating %s fake bans...\n"
        self.stdout.write(message % fake_bans_to_create)

        created_count = 0
        show_progress(self, created_count, fake_bans_to_create)
        for _ in range(fake_bans_to_create):
            ban_faker = random.choice(ban_fakers)
            ban_faker(fake)

            created_count += 1
            show_progress(self, created_count, fake_bans_to_create)

        Ban.objects.invalidate_cache()

        message = "\n\nSuccessfully created %s fake bans"
        self.stdout.write(message % created_count)



from django.contrib import messages
from django.db.models import ObjectDoesNotExist
from django.shortcuts import redirect
from django.utils.translation import pgettext, pgettext_lazy

from ...admin.views import generic
from ..cache import clear_theme_cache
from ..models import Theme, Css
from .css import move_css_down, move_css_up
from .exporter import export_theme
from .forms import (
    CssEditorForm,
    CssLinkForm,
    ImportForm,
    ThemeForm,
    UploadCssForm,
    UploadMediaForm,
)
from .importer import ThemeImportError, import_theme
from .tasks import build_single_theme_css, build_theme_css, update_remote_css_size


class ThemeAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:themes:index"
    model = Theme
    form_class = ThemeForm
    templates_dir = "misago/admin/themes"
    message_404 = pgettext_lazy("admin themes", "Requested theme does not exist.")


class ThemesList(ThemeAdmin, generic.ListView):
    pass


class NewTheme(ThemeAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy(
        "admin themes", 'New theme "%(name)s" has been saved.'
    )

    def get_form(self, form_class, request, target):
        if request.method == "POST":
            return form_class(request.POST, request.FILES, instance=target)

        try:
            initial = {"parent": int(request.GET.get("parent"))}
        except (TypeError, ValueError):
            initial = {}

        return form_class(initial=initial)


class EditTheme(ThemeAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy("admin themes", 'Theme "%(name)s" has been updated.')

    def check_permissions(self, request, target):
        if target.is_default:
            return pgettext("admin themes", "Default theme can't be edited.")

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)
        if "parent" in form.changed_data:
            clear_theme_cache()


class DeleteTheme(ThemeAdmin, generic.ButtonView):
    message_submit = pgettext_lazy("admin themes", 'Theme "%(name)s" has been deleted.')

    def check_permissions(self, request, target):
        if target.is_default:
            return pgettext("admin themes", "Default theme can't be deleted.")
        if target.is_active:
            return pgettext("admin themes", "Active theme can't be deleted.")
        if target.get_descendants().filter(is_active=True).exists():
            message = pgettext(
                "admin themes",
                'Theme "%(name)s" can\'t be deleted because one of its child themes is set as an active theme.',
            )
            return message % {"name": target}

    def button_action(self, request, target):
        for theme in reversed(target.get_descendants(include_self=True)):
            theme.delete()

        clear_theme_cache()
        messages.success(request, self.message_submit % {"name": target})


class ActivateTheme(ThemeAdmin, generic.ButtonView):
    def button_action(self, request, target):
        set_theme_as_active(request, target)

        message = pgettext(
            "admin themes", 'Active theme has been changed to "%(name)s".'
        )
        messages.success(request, message % {"name": target})


def set_theme_as_active(request, theme):
    Theme.objects.update(is_active=False)
    Theme.objects.filter(pk=theme.pk).update(is_active=True)
    clear_theme_cache()


class ExportTheme(ThemeAdmin, generic.ButtonView):
    def check_permissions(self, request, target):
        if target.is_default:
            return pgettext("admin themes", "Default theme can't be exported.")

    def button_action(self, request, target):
        return export_theme(target)


class ImportTheme(ThemeAdmin, generic.FormView):
    form_class = ImportForm
    template_name = "import.html"

    def handle_form(self, form, request):
        try:
            self.import_theme(request, **form.cleaned_data)
            return redirect(self.root_link)
        except ThemeImportError as e:
            form.add_error("upload", str(e))
            return self.render(request, {"form": form})

    def import_theme(self, request, *_, name, parent, upload):
        theme = import_theme(name, parent, upload)
        message = pgettext("admin themes", 'Theme "%(name)s" has been imported.')
        messages.success(request, message % {"name": theme})


class ThemeAssetsAdmin(ThemeAdmin):
    def check_permissions(self, request, theme):
        if theme.is_default:
            return pgettext("admin themes", "Default theme assets can't be edited.")

    def redirect_to_theme_assets(self, theme):
        return redirect("misago:admin:themes:assets", pk=theme.pk)


class ThemeAssets(ThemeAssetsAdmin, generic.TargetedView):
    template_name = "assets/list.html"

    def real_dispatch(self, request, theme):
        return self.render(request, {"theme": theme})


class ThemeAssetsActionAdmin(ThemeAssetsAdmin):
    def real_dispatch(self, request, theme):
        if request.method == "POST":
            self.action(request, theme)

        return self.redirect_to_theme_assets(theme)

    def action(self, request, theme):
        raise NotImplementedError(
            "action method must be implemented in inheriting class"
        )


class UploadThemeAssets(ThemeAssetsActionAdmin, generic.TargetedView):
    message_partial_success = pgettext_lazy(
        "admin themes", "Some css files could not have been added to the theme."
    )

    message_submit = None
    form_class = None

    def action(self, request, theme):
        form = self.form_class(  # pylint: disable=not-callable
            request.POST, request.FILES, instance=theme
        )

        if not form.is_valid():
            if form.cleaned_data.get("assets"):
                messages.info(request, self.message_partial_success)
            for error in form.errors["assets"]:
                messages.error(request, error)

        if form.cleaned_data.get("assets"):
            form.save()
            build_theme_css.delay(theme.pk)
            messages.success(request, self.message_success)


class UploadThemeCss(UploadThemeAssets):
    message_success = pgettext_lazy(
        "admin themes", "New CSS files have been added to the theme."
    )
    form_class = UploadCssForm


class UploadThemeMedia(UploadThemeAssets):
    message_success = pgettext_lazy(
        "admin themes", "New media files have been added to the theme."
    )
    form_class = UploadMediaForm


class DeleteThemeAssets(ThemeAssetsActionAdmin, generic.TargetedView):
    message_submit = None
    queryset_attr = None

    def action(self, request, theme):
        items = self.clean_items_list(request)
        if items:
            queryset = getattr(theme, self.queryset_attr)
            for item in items:
                self.delete_asset(queryset, item)

            messages.success(request, self.message_submit)

    def clean_items_list(self, request):
        try:
            return {int(i) for i in request.POST.getlist("item")[:20]}
        except (ValueError, TypeError):
            pass

    def delete_asset(self, queryset, item):
        try:
            queryset.get(pk=item).delete()
        except ObjectDoesNotExist:
            pass


class DeleteThemeCss(DeleteThemeAssets):
    message_submit = pgettext_lazy(
        "admin themes", "Selected CSS files have been deleted."
    )
    queryset_attr = "css"

    def action(self, request, theme):
        super().action(request, theme)
        clear_theme_cache()


class DeleteThemeMedia(DeleteThemeAssets):
    message_submit = pgettext_lazy("admin themes", "Selected media have been deleted.")
    queryset_attr = "media"


class ThemeCssAdmin(ThemeAssetsAdmin, generic.TargetedView):
    def wrapped_dispatch(self, request, pk, css_pk=None):
        theme = self.get_target_or_none(request, {"pk": pk})
        if not theme:
            messages.error(request, self.message_404)
            return redirect(self.root_link)

        error = self.check_permissions(  # pylint: disable=assignment-from-no-return
            request, theme
        )
        if error:
            messages.error(request, error)
            return redirect(self.root_link)

        css = self.get_theme_css_or_none(theme, css_pk)
        if css_pk and not css:
            css_error = pgettext(
                "admin themes", "Requested CSS could not be found in the theme."
            )
            messages.error(request, css_error)
            return self.redirect_to_theme_assets(theme)

        return self.real_dispatch(request, theme, css)

    def get_theme_css_or_none(self, theme, css_pk):
        if not css_pk:
            return None

        try:
            return theme.css.select_for_update().get(pk=css_pk)
        except ObjectDoesNotExist:
            return None

    def real_dispatch(self, request, theme, css):
        raise NotImplementedError(
            "Admin views extending the ThemeCssAdmin"
            "should define real_dispatch(request, theme, css)"
        )


class MoveThemeCssUp(ThemeCssAdmin):
    def real_dispatch(self, request, theme, css):
        if request.method == "POST" and move_css_up(theme, css):
            clear_theme_cache()
            messages.success(
                request, pgettext("admin themes", '"%s" was moved up.') % css
            )

        return self.redirect_to_theme_assets(theme)


class MoveThemeCssDown(ThemeCssAdmin):
    def real_dispatch(self, request, theme, css):
        if request.method == "POST" and move_css_down(theme, css):
            clear_theme_cache()
            messages.success(
                request, pgettext("admin themes", '"%s" was moved down.') % css
            )

        return self.redirect_to_theme_assets(theme)


class ThemeCssFormAdmin(ThemeCssAdmin, generic.ModelFormView):
    is_atomic = False  # atomic updates cause race condition with celery tasks

    def real_dispatch(self, request, theme, css=None):
        form = self.get_form(self.form_class, request, theme, css)

        if request.method == "POST" and form.is_valid():
            response = self.handle_form(  # pylint: disable=assignment-from-no-return
                form, request, theme, css
            )
            if response:
                return response
            if "stay" in request.POST:
                return self.redirect_to_edit_form(theme, form.instance)
            return self.redirect_to_theme_assets(theme)

        template_name = self.get_template_name(request, css)
        return self.render(
            request, {"form": form, "theme": theme, "target": css}, template_name
        )

    def get_form(self, form_class, request, theme, css):
        raise NotImplementedError(
            "Admin views extending the ThemeCssFormAdmin "
            "should define the get_form(form_class, request, theme, css)"
        )

    def handle_form(self, form, request, theme, css):
        form.save()
        if css.source_needs_building:
            build_single_theme_css.delay(css.pk)
        else:
            clear_theme_cache()
        messages.success(request, self.message_submit % {"name": css.name})


class NewThemeCss(ThemeCssFormAdmin):
    message_submit = pgettext_lazy("admin themes", 'New CSS "%(name)s" has been saved.')
    form_class = CssEditorForm
    template_name = "assets/css-editor-form.html"

    def get_theme_css_or_none(self, theme, _):
        return Css(theme=theme)

    def get_form(self, form_class, request, theme, css):
        if request.method == "POST":
            return form_class(request.POST, instance=css)
        return form_class(instance=css)

    def redirect_to_edit_form(self, theme, css):
        return redirect("misago:admin:themes:edit-css-file", pk=theme.pk, css_pk=css.pk)


class EditThemeCss(NewThemeCss):
    message_submit = pgettext_lazy("admin themes", 'CSS "%(name)s" has been updated.')

    def get_theme_css_or_none(self, theme, css_pk):
        try:
            return theme.css.get(pk=css_pk, url__isnull=True)
        except ObjectDoesNotExist:
            return None

    def get_form(self, form_class, request, theme, css):
        if request.method == "POST":
            return form_class(request.POST, instance=css)
        initial_data = {"source": css.source_file.read().decode("utf-8")}
        return form_class(instance=css, initial=initial_data)

    def handle_form(self, form, request, theme, css):
        if "source" in form.changed_data:
            form.save()
            if css.source_needs_building:
                build_single_theme_css.delay(css.pk)
            else:
                clear_theme_cache()
            messages.success(request, self.message_submit % {"name": css.name})
        else:
            message = pgettext(
                "admin themes", 'No changes have been made to "%(css)s".'
            )
            messages.info(request, message % {"name": css.name})


class NewThemeCssLink(ThemeCssFormAdmin):
    message_submit = pgettext_lazy(
        "admin themes", 'New CSS link "%(name)s" has been saved.'
    )
    form_class = CssLinkForm
    template_name = "assets/css-link-form.html"

    def get_theme_css_or_none(self, theme, _):
        return Css(theme=theme)

    def get_form(self, form_class, request, theme, css):
        if request.method == "POST":
            return form_class(request.POST, instance=css)
        return form_class(instance=css)

    def handle_form(self, form, request, theme, css):
        super().handle_form(form, request, theme, css)
        if "url" in form.changed_data:
            update_remote_css_size.delay(css.pk)
            clear_theme_cache()

    def redirect_to_edit_form(self, theme, css):
        return redirect("misago:admin:themes:new-css-link", pk=theme.pk)


class EditThemeCssLink(NewThemeCssLink):
    message_submit = pgettext_lazy(
        "admin themes", 'CSS link "%(name)s" has been updated.'
    )

    def get_theme_css_or_none(self, theme, css_pk):
        try:
            return theme.css.get(pk=css_pk, url__isnull=False)
        except ObjectDoesNotExist:
            return None

    def redirect_to_edit_form(self, theme, css):
        return redirect("misago:admin:themes:edit-css-link", pk=theme.pk, css_pk=css.pk)



from datetime import timedelta

from django.utils import timezone

from .. import test
from ...categories import THREADS_ROOT_NAME
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..models import Thread
from ..test import patch_category_acl
from ..threadtypes import trees_map


class ThreadsApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        threads_tree_id = trees_map.get_tree_id_for_root(THREADS_ROOT_NAME)

        self.root = Category.objects.get(tree_id=threads_tree_id, level=0)
        self.category = Category.objects.get(slug="first-category")

        self.thread = test.post_thread(category=self.category)
        self.api_link = self.thread.get_api_url()

    def get_thread_json(self):
        response = self.client.get(self.thread.get_api_url())
        self.assertEqual(response.status_code, 200)

        return response.json()


class ThreadRetrieveApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.tested_links = [
            self.api_link,
            "%sposts/" % self.api_link,
            "%sposts/?page=1" % self.api_link,
        ]

    @patch_category_acl()
    def test_api_returns_thread(self):
        """api has no showstoppers"""
        for link in self.tested_links:
            response = self.client.get(link)
            self.assertEqual(response.status_code, 200)

            response_json = response.json()
            self.assertEqual(response_json["id"], self.thread.pk)
            self.assertEqual(response_json["title"], self.thread.title)

            if "posts" in link:
                self.assertIn("post_set", response_json)

    @patch_category_acl({"can_see_all_threads": False})
    def test_api_shows_owned_thread(self):
        """api handles "owned threads only"""
        for link in self.tested_links:
            response = self.client.get(link)
            self.assertEqual(response.status_code, 404)

        self.thread.starter = self.user
        self.thread.save()

        for link in self.tested_links:
            response = self.client.get(link)
            self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_see": False})
    def test_api_validates_category_see_permission(self):
        """api validates category visiblity"""
        for link in self.tested_links:
            response = self.client.get(link)
            self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_browse": False})
    def test_api_validates_category_browse_permission(self):
        """api validates category browsability"""
        for link in self.tested_links:
            response = self.client.get(link)
            self.assertEqual(response.status_code, 404)

    def test_api_validates_posts_visibility(self):
        """api validates posts visiblity"""
        hidden_post = test.reply_thread(
            self.thread, is_hidden=True, message="I'am hidden test message!"
        )

        with patch_category_acl({"can_hide_posts": 0}):
            response = self.client.get(self.tested_links[1])
            self.assertNotContains(
                response, hidden_post.parsed
            )  # post's body is hidden

        # add permission to see hidden posts
        with patch_category_acl({"can_hide_posts": 1}):
            response = self.client.get(self.tested_links[1])
            self.assertContains(
                response, hidden_post.parsed
            )  # hidden post's body is visible with permission

        # unapproved posts shouldn't show at all
        unapproved_post = test.reply_thread(self.thread, is_unapproved=True)

        with patch_category_acl({"can_approve_content": False}):
            response = self.client.get(self.tested_links[1])
            self.assertNotContains(response, unapproved_post.get_absolute_url())

        # add permission to see unapproved posts
        with patch_category_acl({"can_approve_content": True}):
            response = self.client.get(self.tested_links[1])
            self.assertContains(response, unapproved_post.get_absolute_url())

    def test_api_validates_has_unapproved_posts_visibility(self):
        """api checks acl before exposing unapproved flag"""
        self.thread.has_unapproved_posts = True
        self.thread.save()

        with patch_category_acl({"can_approve_content": False}):
            for link in self.tested_links:
                response = self.client.get(link)
                self.assertEqual(response.status_code, 200)

                response_json = response.json()
                self.assertEqual(response_json["id"], self.thread.pk)
                self.assertEqual(response_json["title"], self.thread.title)
                self.assertFalse(response_json["has_unapproved_posts"])

        with patch_category_acl({"can_approve_content": True}):
            for link in self.tested_links:
                response = self.client.get(link)
                self.assertEqual(response.status_code, 200)

                response_json = response.json()
                self.assertEqual(response_json["id"], self.thread.pk)
                self.assertEqual(response_json["title"], self.thread.title)
                self.assertTrue(response_json["has_unapproved_posts"])


class ThreadDeleteApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.last_thread = test.post_thread(category=self.category)
        self.api_link = self.last_thread.get_api_url()

    def test_delete_thread_no_permission(self):
        """api tests permission to delete threads"""
        with patch_category_acl({"can_hide_threads": 0}):
            response = self.client.delete(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json()["detail"], "You can't delete threads in this category."
            )

        with patch_category_acl({"can_hide_threads": 1}):
            response = self.client.delete(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json()["detail"], "You can't delete threads in this category."
            )

    @patch_category_acl({"can_hide_threads": 1, "can_hide_own_threads": 2})
    def test_delete_other_user_thread_no_permission(self):
        """api tests thread owner when deleting own thread"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"],
            "You can't delete other users theads in this category.",
        )

    @patch_category_acl(
        {"can_hide_threads": 2, "can_hide_own_threads": 2, "can_close_threads": False}
    )
    def test_delete_thread_closed_category_no_permission(self):
        """api tests category's closed state"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"],
            "This category is closed. You can't delete threads in it.",
        )

    @patch_category_acl(
        {"can_hide_threads": 2, "can_hide_own_threads": 2, "can_close_threads": False}
    )
    def test_delete_thread_closed_no_permission(self):
        """api tests thread's closed state"""
        self.last_thread.is_closed = True
        self.last_thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"], "This thread is closed. You can't delete it."
        )

    @patch_category_acl(
        {"can_hide_threads": 1, "can_hide_own_threads": 2, "thread_edit_time": 1}
    )
    def test_delete_owned_thread_no_time(self):
        """api tests permission to delete owned thread within time limit"""
        self.last_thread.starter = self.user
        self.last_thread.started_on = timezone.now() - timedelta(minutes=10)
        self.last_thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"],
            "You can't delete threads that are older than 1 minute.",
        )

    @patch_category_acl({"can_hide_threads": 2})
    def test_delete_thread(self):
        """DELETE to API link with permission deletes thread"""
        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.last_thread_id, self.last_thread.pk)

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.last_thread.pk)

        # category was synchronised after deletion
        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.last_thread_id, self.thread.pk)

        # test that last thread's deletion triggers category sync
        response = self.client.delete(self.thread.get_api_url())
        self.assertEqual(response.status_code, 200)

        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        category = Category.objects.get(slug="first-category")
        self.assertIsNone(category.last_thread_id)

    @patch_category_acl(
        {"can_hide_threads": 1, "can_hide_own_threads": 2, "thread_edit_time": 30}
    )
    def test_delete_owned_thread(self):
        """api lets owner to delete owned thread within time limit"""
        self.last_thread.starter = self.user
        self.last_thread.started_on = timezone.now() - timedelta(minutes=10)
        self.last_thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)



from django.contrib.auth import get_user_model
from django.db import transaction
from django.db.models import Q
from django.db.models.signals import pre_delete
from django.dispatch import Signal, receiver
from django.utils.translation import pgettext

from ..categories.models import Category
from ..categories.signals import delete_category_content, move_category_content
from ..core.pgutils import chunk_queryset
from ..notifications.models import Notification, WatchedThread
from ..users.signals import (
    anonymize_user_data,
    archive_user_data,
    delete_user_content,
    username_changed,
)
from .anonymize import ANONYMIZABLE_EVENTS, anonymize_event, anonymize_post_last_likes
from .models import Attachment, Poll, PollVote, Post, PostEdit, PostLike, Thread

delete_post = Signal()
delete_thread = Signal()
merge_post = Signal()
merge_thread = Signal()
move_post = Signal()
move_thread = Signal()
update_thread_title = Signal()


@receiver(merge_thread)
def merge_threads(sender, **kwargs):
    other_thread = kwargs["other_thread"]

    other_thread.post_set.update(category=sender.category, thread=sender)
    other_thread.postedit_set.update(category=sender.category, thread=sender)
    other_thread.postlike_set.update(category=sender.category, thread=sender)

    other_thread.subscription_set.exclude(
        user__in=sender.subscription_set.values("user")
    ).update(category=sender.category, thread=sender)

    other_thread.notification_set.update(
        category=sender.category,
        thread=sender,
        thread_title=sender.title,
    )

    other_thread.watchedthread_set.update(
        category=sender.category,
        thread=sender,
    )


@receiver(merge_post)
def merge_posts(sender, **kwargs):
    other_post = kwargs["other_post"]
    for user in sender.mentions.iterator():
        other_post.mentions.add(user)

    sender.notification_set.update(post=other_post)


@receiver(move_post)
def move_post_notifications(sender, **kwargs):
    sender.notification_set.update(
        category=sender.category,
        thread=sender.thread,
        thread_title=sender.thread.title,
    )


@receiver(move_thread)
def move_thread_content(sender, **kwargs):
    sender.post_set.update(category=sender.category)
    sender.postedit_set.update(category=sender.category)
    sender.postlike_set.update(category=sender.category)
    sender.pollvote_set.update(category=sender.category)
    sender.subscription_set.update(category=sender.category)
    sender.notification_set.update(category=sender.category)
    sender.watchedthread_set.update(category=sender.category)

    Poll.objects.filter(thread=sender).update(category=sender.category)


@receiver(update_thread_title)
def change_thread_title(sender, **kwargs):
    sender.notification_set.update(thread_title=sender.title)


@receiver(delete_category_content)
def delete_category_threads(sender, **kwargs):
    sender.notification_set.all().delete()
    sender.watchedthread_set.all().delete()
    sender.subscription_set.all().delete()
    sender.pollvote_set.all().delete()
    sender.poll_set.all().delete()
    sender.postlike_set.all().delete()
    sender.thread_set.all().delete()
    sender.postedit_set.all().delete()
    sender.post_set.all().delete()


@receiver(move_category_content)
def move_category_threads(sender, **kwargs):
    new_category = kwargs["new_category"]

    sender.thread_set.update(category=new_category)
    sender.post_set.filter(category=sender).update(category=new_category)
    sender.postedit_set.filter(category=sender).update(category=new_category)
    sender.postlike_set.filter(category=sender).update(category=new_category)
    sender.poll_set.filter(category=sender).update(category=new_category)
    sender.pollvote_set.update(category=new_category)
    sender.subscription_set.update(category=new_category)
    sender.notification_set.update(category=new_category)
    sender.watchedthread_set.update(category=new_category)


@receiver(delete_user_content)
def delete_user_threads(sender, **kwargs):
    recount_categories = set()
    recount_threads = set()

    Notification.objects.filter(
        Q(thread__starter=sender) | Q(post__poster=sender)
    ).delete()

    WatchedThread.objects.filter(thread__starter=sender).delete()

    for post in chunk_queryset(sender.liked_post_set):
        cleaned_likes = list(filter(lambda i: i["id"] != sender.id, post.last_likes))
        if cleaned_likes != post.last_likes:
            post.last_likes = cleaned_likes
            post.save(update_fields=["last_likes"])

    for thread in chunk_queryset(sender.thread_set):
        recount_categories.add(thread.category_id)
        with transaction.atomic():
            thread.delete()

    for post in chunk_queryset(sender.post_set):
        recount_categories.add(post.category_id)
        recount_threads.add(post.thread_id)
        with transaction.atomic():
            post.delete()

    if recount_threads:
        changed_threads_qs = Thread.objects.filter(id__in=recount_threads)
        for thread in chunk_queryset(changed_threads_qs):
            thread.synchronize()
            thread.save()

    if recount_categories:
        for category in Category.objects.filter(id__in=recount_categories):
            category.synchronize()
            category.save()


@receiver(archive_user_data)
def archive_user_attachments(sender, archive=None, **kwargs):
    queryset = sender.attachment_set.order_by("id")
    for attachment in chunk_queryset(queryset):
        archive.add_model_file(
            attachment.file,
            prefix=attachment.uploaded_on.strftime("%H%M%S-file"),
            date=attachment.uploaded_on,
        )
        archive.add_model_file(
            attachment.image,
            prefix=attachment.uploaded_on.strftime("%H%M%S-image"),
            date=attachment.uploaded_on,
        )
        archive.add_model_file(
            attachment.thumbnail,
            prefix=attachment.uploaded_on.strftime("%H%M%S-thumbnail"),
            date=attachment.uploaded_on,
        )


@receiver(archive_user_data)
def archive_user_posts(sender, archive=None, **kwargs):
    queryset = sender.post_set.order_by("id")
    for post in chunk_queryset(queryset):
        item_name = post.posted_on.strftime("%H%M%S-post")
        archive.add_text(item_name, post.parsed, date=post.posted_on)


@receiver(archive_user_data)
def archive_user_posts_edits(sender, archive=None, **kwargs):
    queryset = PostEdit.objects.filter(post__poster=sender).order_by("id")
    for post_edit in chunk_queryset(queryset):
        item_name = post_edit.edited_on.strftime("%H%M%S-post-edit")
        archive.add_text(item_name, post_edit.edited_from, date=post_edit.edited_on)
    queryset = sender.postedit_set.exclude(id__in=queryset.values("id")).order_by("id")
    for post_edit in chunk_queryset(queryset):
        item_name = post_edit.edited_on.strftime("%H%M%S-post-edit")
        archive.add_text(item_name, post_edit.edited_from, date=post_edit.edited_on)


@receiver(archive_user_data)
def archive_user_polls(sender, archive=None, **kwargs):
    queryset = sender.poll_set.order_by("id")
    for poll in chunk_queryset(queryset):
        item_name = poll.posted_on.strftime("%H%M%S-poll")
        archive.add_dict(
            item_name,
            {
                pgettext("archived poll", "Question"): poll.question,
                pgettext("archived poll", "Choices"): ", ".join(
                    [c["label"] for c in poll.choices]
                ),
            },
            date=poll.posted_on,
        )


@receiver(anonymize_user_data)
def anonymize_user_in_events(sender, **kwargs):
    queryset = Post.objects.filter(
        is_event=True,
        event_type__in=ANONYMIZABLE_EVENTS,
        event_context__user__id=sender.id,
    )

    for event in chunk_queryset(queryset):
        anonymize_event(sender, event)


@receiver([anonymize_user_data])
def anonymize_user_in_likes(sender, **kwargs):
    for post in chunk_queryset(sender.liked_post_set):
        anonymize_post_last_likes(sender, post)


@receiver([anonymize_user_data, username_changed])
def update_usernames(sender, **kwargs):
    Thread.objects.filter(starter=sender).update(
        starter_name=sender.username, starter_slug=sender.slug
    )

    Thread.objects.filter(last_poster=sender).update(
        last_poster_name=sender.username, last_poster_slug=sender.slug
    )

    Thread.objects.filter(best_answer_marked_by=sender).update(
        best_answer_marked_by_name=sender.username,
        best_answer_marked_by_slug=sender.slug,
    )

    Post.objects.filter(poster=sender).update(poster_name=sender.username)

    Post.objects.filter(last_editor=sender).update(
        last_editor_name=sender.username, last_editor_slug=sender.slug
    )

    PostEdit.objects.filter(editor=sender).update(
        editor_name=sender.username, editor_slug=sender.slug
    )

    PostLike.objects.filter(liker=sender).update(
        liker_name=sender.username, liker_slug=sender.slug
    )

    Attachment.objects.filter(uploader=sender).update(
        uploader_name=sender.username, uploader_slug=sender.slug
    )

    Poll.objects.filter(poster=sender).update(
        poster_name=sender.username, poster_slug=sender.slug
    )

    PollVote.objects.filter(voter=sender).update(
        voter_name=sender.username, voter_slug=sender.slug
    )


@receiver(pre_delete, sender=get_user_model())
def remove_unparticipated_private_threads(sender, **kwargs):
    threads_qs = kwargs["instance"].privatethread_set.all()
    for thread in chunk_queryset(threads_qs):
        if thread.participants.count() == 1:
            with transaction.atomic():
                thread.delete()



from django.core import mail
from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ..test import AuthenticatedUserTestCase


class UserChangePasswordTests(AuthenticatedUserTestCase):
    """tests for user change password RPC (/api/users/1/change-password/)"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/change-password/" % self.user.pk

    def test_unsupported_methods(self):
        """api isn't supporting GET"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 405)

    def test_empty_input(self):
        """api errors correctly for empty input"""
        response = self.client.post(self.link, data={})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "new_password": ["This field is required."],
                "password": ["This field is required."],
            },
        )

    def test_invalid_password(self):
        """api errors correctly for invalid password"""
        response = self.client.post(
            self.link, data={"new_password": "N3wP@55w0rd", "password": "Lor3mIpsum"}
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"password": ["Entered password is invalid."]}
        )

    def test_blank_input(self):
        """api errors correctly for blank input"""
        response = self.client.post(
            self.link, data={"new_password": "", "password": self.USER_PASSWORD}
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"new_password": ["This field may not be blank."]}
        )

    def test_short_new_pasword(self):
        """api errors correctly for short new password"""
        response = self.client.post(
            self.link, data={"new_password": "n", "password": self.USER_PASSWORD}
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "new_password": [
                    "This password is too short. It must contain at least 7 characters."
                ]
            },
        )

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_change_password(self):
        """api allows users to change their passwords"""
        new_password = "N3wP@55w0rd"

        response = self.client.post(
            self.link,
            data={"new_password": new_password, "password": self.USER_PASSWORD},
        )
        self.assertEqual(response.status_code, 200)

        self.assertIn("Confirm password change", mail.outbox[0].subject)
        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                token = line.rstrip("/").split("/")[-1]
                break
        else:
            self.fail("E-mail sent didn't contain confirmation url")

        response = self.client.get(
            reverse("misago:options-confirm-password-change", kwargs={"token": token})
        )

        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertTrue(self.user.check_password(new_password))

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_change_password_with_whitespaces(self):
        """api handles users with whitespaces around their passwords"""
        old_password = " old password "
        new_password = " N3wP@55w0rd "

        self.user.set_password(old_password)
        self.user.save()

        self.login_user(self.user)

        response = self.client.post(
            self.link, data={"new_password": new_password, "password": old_password}
        )
        self.assertEqual(response.status_code, 200)

        self.assertIn("Confirm password change", mail.outbox[0].subject)
        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                token = line.rstrip("/").split("/")[-1]
                break
        else:
            self.fail("E-mail sent didn't contain confirmation url")

        response = self.client.get(
            reverse("misago:options-confirm-password-change", kwargs={"token": token})
        )

        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertTrue(self.user.check_password(new_password))

    @override_dynamic_settings(
        enable_oauth2_client=True,
        oauth2_provider="Lorem",
    )
    def test_change_password_api_returns_403_if_oauth_is_enabled(self):
        new_password = " N3wP@55w0rd "

        self.login_user(self.user)

        response = self.client.post(
            self.link,
            data={"new_password": new_password, "password": self.USER_PASSWORD},
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(len(mail.outbox), 0)

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_confirm_change_password_view_returns_403_if_oauth_is_enabled(self):
        new_password = " N3wP@55w0rd "

        self.login_user(self.user)

        response = self.client.post(
            self.link,
            data={"new_password": new_password, "password": self.USER_PASSWORD},
        )
        self.assertEqual(response.status_code, 200)

        self.assertIn("Confirm password change", mail.outbox[0].subject)
        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                token = line.rstrip("/").split("/")[-1]
                break
        else:
            self.fail("E-mail sent didn't contain confirmation url")

        with override_dynamic_settings(
            enable_oauth2_client=True, oauth2_provider="Lorem"
        ):
            response = self.client.get(
                reverse(
                    "misago:options-confirm-password-change", kwargs={"token": token}
                )
            )

            self.assertEqual(response.status_code, 403)



from ..site import Node


def test_node_is_added_at_end_of_parent_children():
    master = Node(name="Apples", link="misago:index")
    child = Node(name="Oranges", link="misago:index")
    master.add_node(child)

    assert master.children()[-1].name == child.name


def test_add_node_after():
    """add_node added node after specific node"""
    master = Node(name="Apples", link="misago:index")

    child = Node(name="Oranges", link="misago:index")
    master.add_node(child)

    test = Node(name="Potatoes", link="misago:index")
    master.add_node(test, after="misago:index")

    all_nodes = master.children()
    for i, node in enumerate(all_nodes):
        if node.name == test.name:
            assert all_nodes[i - 1].name == child.name


def test_add_node_before():
    """add_node added node  before specific node"""
    master = Node(name="Apples", link="misago:index")

    child = Node(name="Oranges", link="misago:index")
    master.add_node(child)

    test = Node(name="Potatoes", link="misago:index")
    master.add_node(test, before="misago:index")

    all_nodes = master.children()
    for i, node in enumerate(all_nodes):
        if node.name == test.name:
            assert all_nodes[i + 1].name == child.name



from unittest.mock import patch

import pytest


def noop_filter_user_data(request, user, user_data):
    return user_data


@pytest.fixture
def disable_user_data_filters():
    with patch("misago.oauth2.validation.filter_user_data", noop_filter_user_data):
        yield



from django.http import HttpRequest
from django.shortcuts import render

from ..readtracker.categories import get_categories_new_posts
from ..users.models import User
from .enums import CategoryTree
from .models import Category


def index(request):
    categories_list = get_categories_list(request)

    return render(
        request,
        "misago/categories/index.html",
        {"categories_list": categories_list},
    )


def get_categories_list(request: HttpRequest):
    if not request.user_acl["visible_categories"]:
        return []

    categories_qs = Category.objects.filter(
        id__in=request.user_acl["visible_categories"],
        tree_id=CategoryTree.THREADS,
        level__gt=0,
    )

    categories_map: dict[int, dict] = {}

    for category in categories_qs:
        category_permissions = request.user_acl["categories"][category.id]

        if category.last_thread_id:
            category_last_thread = {
                "id": category.last_thread_id,
                "title": category.last_thread_title,
                "slug": category.last_thread_slug,
                "last_post_on": category.last_post_on,
                "last_poster": None,
                "last_poster_name": category.last_poster_name,
            }
        else:
            category_last_thread = None

        categories_map[category.id] = {
            "category": category,
            "threads": category.threads,
            "posts": category.posts,
            "last_thread": category_last_thread,
            "new_posts": False,
            "is_protected": category.id
            not in request.user_acl["browseable_categories"],
            "is_private": not category_permissions["can_see_all_threads"],
            "children": [],
            "children_threads": category.threads,
            "children_posts": category.posts,
            "children_last_thread": category_last_thread,
            "children_new_posts": False,
        }

    last_posters = prefetch_last_posters(request, categories_map.values())
    new_posts: dict[int, bool] = {}

    if request.user.is_authenticated:
        new_posts = get_categories_new_posts(
            request,
            [item["category"] for item in categories_map.values()],
        )

    # Populate categories last posters and read states
    # Aggregate categories to their parents
    for item in reversed(categories_map.values()):
        category = item["category"]

        if category.last_poster_id:
            last_poster = last_posters[category.last_poster_id]
            item["last_thread"].update(
                {
                    "last_poster": last_poster,
                    "last_poster_name": last_poster.username,
                }
            )

        if request.user.is_authenticated and new_posts[category.id]:
            item["new_posts"] = True
            item["children_new_posts"] = True

        if category.level > 1:
            parent = categories_map[category.parent_id]

            # Add item's aggregated stats to parent's
            parent["children_threads"] += item["children_threads"]
            parent["children_posts"] += item["children_posts"]

            # Update parent's last thread if they don't have one or its older
            item_last_thread = item["children_last_thread"]
            parent_last_thread = parent["children_last_thread"]
            if item_last_thread and (
                not parent_last_thread
                or item_last_thread["last_post_on"] > parent_last_thread["last_post_on"]
            ):
                parent["children_last_thread"] = item_last_thread

            # Propagate to parent the new posts status
            if item["children_new_posts"]:
                parent["children_new_posts"] = True

            parent["children"].insert(0, item)

    return [
        category
        for category in categories_map.values()
        if category["category"].level == 1
    ]


def prefetch_last_posters(
    request: HttpRequest, categories_list: list[dict]
) -> dict[int, User]:
    if not categories_list:
        return []

    last_posters: dict[int, User] = {}
    last_posters_ids: set[int] = set()
    for item in categories_list:
        last_poster_id = item["category"].last_poster_id
        if last_poster_id:
            if last_poster_id == request.user.id:
                last_posters[request.user.id] = request.user
            else:
                last_posters_ids.add(last_poster_id)

    if last_posters_ids:
        users_qs = User.objects.filter(id__in=last_posters_ids).select_related("rank")
        for user in users_qs:
            last_posters[user.id] = user

    return last_posters



from django.contrib import messages
from django.shortcuts import redirect, render
from django.utils.translation import pgettext
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters

from .. import auth
from ...users.forms.auth import AdminAuthenticationForm


@sensitive_post_parameters()
@csrf_protect
@never_cache
def login(request):
    if request.admin_namespace == "misago:admin":
        target = "misago"
    elif request.admin_namespace == "admin":
        target = "django"
    else:
        target = "unknown"

    form = AdminAuthenticationForm(request)

    if request.method == "POST":
        form = AdminAuthenticationForm(request, data=request.POST)
        if form.is_valid():
            auth.login(request, form.user_cache)
            return redirect("%s:index" % request.admin_namespace)

    return render(request, "misago/admin/login.html", {"form": form, "target": target})


@csrf_protect
@never_cache
def logout(request):
    if request.method == "POST":
        auth.remove_admin_authorization(request)
        messages.info(
            request, pgettext("admin auth", "Your admin session has been closed.")
        )
    return redirect("misago:admin:index")



from datetime import timedelta

from ..namechanges import get_left_namechanges, get_next_available_namechange


def test_user_without_permission_to_change_name_has_no_changes_left(user):
    user_acl = {"name_changes_allowed": 0}
    assert get_left_namechanges(user, user_acl) == 0


def test_user_without_namechanges_has_all_changes_left(user):
    user_acl = {"name_changes_allowed": 3, "name_changes_expire": 0}
    assert get_left_namechanges(user, user_acl) == 3


def test_user_own_namechanges_are_subtracted_from_changes_left(user):
    user_acl = {"name_changes_allowed": 3, "name_changes_expire": 0}
    user.set_username("Changed")
    assert get_left_namechanges(user, user_acl) == 2


def test_user_own_recent_namechanges_subtract_from_changes_left(user):
    user_acl = {"name_changes_allowed": 3, "name_changes_expire": 5}
    user.set_username("Changed")
    assert get_left_namechanges(user, user_acl) == 2


def test_user_own_expired_namechanges_dont_subtract_from_changes_left(user):
    user_acl = {"name_changes_allowed": 3, "name_changes_expire": 5}

    username_change = user.set_username("Changed")
    username_change.changed_on -= timedelta(days=10)
    username_change.save()

    assert get_left_namechanges(user, user_acl) == 3


def test_user_namechanges_by_other_users_dont_subtract_from_changes_left(user):
    user_acl = {"name_changes_allowed": 3, "name_changes_expire": 0}

    username_change = user.set_username("Changed")
    username_change.changed_by = None
    username_change.save()

    assert get_left_namechanges(user, user_acl) == 3


def test_user_next_available_namechange_is_none_for_user_with_changes_left(user):
    user_acl = {"name_changes_allowed": 3, "name_changes_expire": 0}
    assert get_next_available_namechange(user, user_acl, 3) is None


def test_user_next_available_namechange_is_none_if_own_namechanges_dont_expire(user):
    user_acl = {"name_changes_allowed": 1, "name_changes_expire": 0}
    user.set_username("Changed")
    assert get_next_available_namechange(user, user_acl, 0) is None


def test_user_next_available_namechange_is_calculated_if_own_namechanges_expire(user):
    user_acl = {"name_changes_allowed": 1, "name_changes_expire": 1}

    username_change = user.set_username("Changed")
    next_change_on = get_next_available_namechange(user, user_acl, 0)

    assert next_change_on
    assert next_change_on == username_change.changed_on + timedelta(days=1)



from django.urls import reverse

from .. import test
from ..serializers import PostLikeSerializer
from ..test import patch_category_acl
from .test_threads_api import ThreadsApiTestCase


class ThreadPostLikesApiTestCase(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.post = test.reply_thread(self.thread, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-post-likes",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    @patch_category_acl({"can_see_posts_likes": 0})
    def test_no_permission(self):
        """api errors if user has no permission to see likes"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't see who liked this post."}
        )

    @patch_category_acl({"can_see_posts_likes": 1})
    def test_no_permission_to_list(self):
        """api errors if user has no permission to see likes, but can see likes count"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't see who liked this post."}
        )

    @patch_category_acl({"can_see_posts_likes": 2})
    def test_no_likes(self):
        """api returns empty list if post has no likes"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])

    @patch_category_acl({"can_see_posts_likes": 2})
    def test_likes(self):
        """api returns list of likes"""
        like = test.like_post(self.post, self.user)
        other_like = test.like_post(self.post, self.user)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                PostLikeSerializer(
                    {
                        "id": other_like.id,
                        "liked_on": other_like.liked_on,
                        "liker_id": other_like.liker_id,
                        "liker_name": other_like.liker_name,
                        "liker_slug": other_like.liker_slug,
                        "liker__avatars": self.user.avatars,
                    }
                ).data,
                PostLikeSerializer(
                    {
                        "id": like.id,
                        "liked_on": like.liked_on,
                        "liker_id": like.liker_id,
                        "liker_name": like.liker_name,
                        "liker_slug": like.liker_slug,
                        "liker__avatars": self.user.avatars,
                    }
                ).data,
            ],
        )

        # api has no showstoppers for likes by deleted users
        like.liker = None
        like.save()

        other_like.liker = None
        other_like.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                PostLikeSerializer(
                    {
                        "id": other_like.id,
                        "liked_on": other_like.liked_on,
                        "liker_id": other_like.liker_id,
                        "liker_name": other_like.liker_name,
                        "liker_slug": other_like.liker_slug,
                        "liker__avatars": None,
                    }
                ).data,
                PostLikeSerializer(
                    {
                        "id": like.id,
                        "liked_on": like.liked_on,
                        "liker_id": like.liker_id,
                        "liker_name": like.liker_name,
                        "liker_slug": like.liker_slug,
                        "liker__avatars": None,
                    }
                ).data,
            ],
        )



from ...acl.objectacl import add_acl_to_obj
from ...core.shortcuts import paginate, pagination_dict
from ...readtracker.poststracker import make_read_aware
from ...users.online.utils import make_users_status_aware
from ..paginator import PostsPaginator
from ..permissions import exclude_invisible_posts
from ..serializers import PostSerializer
from ..utils import add_likes_to_posts

__all__ = ["ThreadPosts"]


class ViewModel:
    def __init__(self, request, thread, page):  # pylint: disable=too-many-locals
        try:
            thread_model = thread.unwrap()
        except AttributeError:
            thread_model = thread

        posts_queryset = self.get_posts_queryset(request, thread_model)

        posts_limit = request.settings.posts_per_page
        posts_orphans = request.settings.posts_per_page_orphans
        list_page = paginate(
            posts_queryset, page, posts_limit, posts_orphans, paginator=PostsPaginator
        )
        paginator = pagination_dict(list_page)

        posts = list(list_page.object_list)
        posters = []

        for post in posts:
            post.category = thread.category
            post.thread = thread_model

            if post.poster:
                posters.append(post.poster)

        make_users_status_aware(request, posters)

        if thread.category.acl["can_see_posts_likes"]:
            add_likes_to_posts(request.user, posts)

        # add events to posts
        if thread_model.has_events:
            first_post = None
            if list_page.has_previous():
                first_post = posts[0]
            last_post = None
            if list_page.has_next():
                last_post = posts[-1]

            events_limit = request.settings.events_per_page
            posts += self.get_events_queryset(
                request, thread_model, events_limit, first_post, last_post
            )

            # sort both by pk
            posts.sort(key=lambda p: p.pk)

        # make posts and events ACL and reads aware
        add_acl_to_obj(request.user_acl, posts)
        make_read_aware(request, posts)

        self._user = request.user

        self.posts = posts
        self.paginator = paginator

    def get_posts_queryset(self, request, thread):
        queryset = (
            thread.post_set.select_related(
                "category",
                "poster",
                "poster__rank",
                "poster__ban_cache",
                "poster__online_tracker",
            )
            .filter(is_event=False)
            .order_by("id")
        )
        return exclude_invisible_posts(request.user_acl, thread.category, queryset)

    def get_events_queryset(
        self, request, thread, limit, first_post=None, last_post=None
    ):
        queryset = thread.post_set.select_related(
            "category",
            "poster",
            "poster__rank",
            "poster__ban_cache",
            "poster__online_tracker",
        ).filter(is_event=True)

        if first_post:
            queryset = queryset.filter(pk__gt=first_post.pk)
        if last_post:
            queryset = queryset.filter(pk__lt=last_post.pk)

        queryset = exclude_invisible_posts(request.user_acl, thread.category, queryset)
        return list(queryset.order_by("-id")[:limit])

    def get_frontend_context(self):
        context = {
            "results": PostSerializer(
                self.posts, many=True, context={"user": self._user}
            ).data
        }

        context.update(self.paginator)

        return context

    def get_template_context(self):
        return {"posts": self.posts, "paginator": self.paginator}


class ThreadPosts(ViewModel):
    pass



from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.test import Client, TestCase, override_settings
from django.test.client import RequestFactory
from django.urls import reverse

from ...acl.useracl import get_user_acl
from ...conf.dynamicsettings import DynamicSettings
from ...conftest import get_cache_versions
from ...users.models import AnonymousUser
from ..testproject.views import mock_custom_403_error_page, mock_custom_404_error_page
from ..utils import encode_json_html


class CSRFErrorViewTests(TestCase):
    def test_csrf_failure_is_handled(self):
        """csrf_failure error page has no show-stoppers"""
        csrf_client = Client(enforce_csrf_checks=True)
        response = csrf_client.post(reverse("misago:index"), data={"eric": "fish"})
        self.assertContains(response, "Request blocked", status_code=403)

    def test_ajax_csrf_failure_is_handled(self):
        """csrf_failure error ajax response has no show-stoppers"""
        csrf_client = Client(enforce_csrf_checks=True)
        response = csrf_client.post(
            reverse("misago:api:auth"),
            data={"eric": "fish"},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "CSRF cookie", status_code=403)


@override_settings(ROOT_URLCONF="misago.core.testproject.urls")
class ErrorPageViewsTests(TestCase):
    def test_banned_returns_403(self):
        """banned error page has no show-stoppers"""
        response = self.client.get(reverse("raise-misago-banned"))
        self.assertContains(response, "misago:error-banned", status_code=403)
        self.assertContains(response, "<p>Banned for test!</p>", status_code=403)
        self.assertContains(
            response, encode_json_html("<p>Banned for test!</p>"), status_code=403
        )

    def test_permission_denied_returns_403(self):
        """permission_denied error page has no show-stoppers"""
        response = self.client.get(reverse("raise-misago-403"))
        self.assertContains(response, "misago:error-403", status_code=403)
        self.assertContains(response, "Page not available", status_code=403)

    def test_page_not_found_returns_404(self):
        """page_not_found error page has no show-stoppers"""
        response = self.client.get(reverse("raise-misago-404"))
        self.assertContains(response, "misago:error-404", status_code=404)
        self.assertContains(response, "Page not found", status_code=404)

    def test_not_allowed_returns_405(self):
        """not allowed error page has no showstoppers"""
        response = self.client.get(reverse("raise-misago-405"))
        self.assertContains(response, "misago:error-405", status_code=405)
        self.assertContains(response, "Wrong way", status_code=405)

    def test_social_auth_failed_returns_403(self):
        """social auth's failed error returns 403"""
        response = self.client.get(reverse("raise-social-auth-failed"))
        self.assertContains(response, "page-error-social", status_code=403)
        self.assertContains(response, "GitHub", status_code=403)

    def test_social_wrong_backend_returns_403(self):
        """social auth's wrong backend error returns 403"""
        response = self.client.get(reverse("raise-social-wrong-backend"))
        self.assertContains(response, "page-error-social", status_code=403)

    def test_social_not_allowed_to_disconnect_returns_403(self):
        """social auth's not allowed to disconnect error returns 403"""
        response = self.client.get(reverse("raise-social-not-allowed-to-disconnect"))
        self.assertContains(response, "page-error-social", status_code=403)

    def test_social_failed_message(self):
        """misago-specific social auth failed error page returns 403 with message"""
        response = self.client.get(reverse("raise-social-auth-failed-message"))
        self.assertContains(response, "page-error-social", status_code=403)
        self.assertContains(
            response, "This message will be shown to user!", status_code=403
        )

    def test_social_auth_banned(self):
        """misago-specific social auth banned error page returns 403 with ban message"""
        response = self.client.get(reverse("raise-social-auth-banned"))
        self.assertContains(response, "page-error-social", status_code=403)
        self.assertContains(response, "Banned in auth!", status_code=403)


def create_request(url):
    request = RequestFactory().get(url)
    request.cache_versions = get_cache_versions()
    request.settings = DynamicSettings(request.cache_versions)
    request.user = AnonymousUser()
    request.user_acl = get_user_acl(request.user, request.cache_versions)
    request.include_frontend_context = True
    request.frontend_context = {}
    request.socialauth = {}
    return request


@override_settings(ROOT_URLCONF="misago.core.testproject.urlswitherrorhandlers")
class CustomErrorPagesTests(TestCase):
    def setUp(self):
        self.misago_request = create_request(reverse("misago:index"))
        self.site_request = create_request(reverse("raise-403"))

    def test_shared_403_decorator(self):
        """shared_403_decorator calls correct error handler"""
        response = self.client.get(reverse("raise-misago-403"))
        self.assertEqual(response.status_code, 403)
        response = self.client.get(reverse("raise-403"))
        self.assertContains(response, "Custom 403", status_code=403)

        response = mock_custom_403_error_page(self.misago_request, PermissionDenied())
        self.assertNotContains(response, "Custom 403", status_code=403)
        response = mock_custom_403_error_page(self.site_request, PermissionDenied())
        self.assertContains(response, "Custom 403", status_code=403)

    def test_shared_404_decorator(self):
        """shared_404_decorator calls correct error handler"""
        response = self.client.get(reverse("raise-misago-404"))
        self.assertEqual(response.status_code, 404)
        response = self.client.get(reverse("raise-404"))
        self.assertContains(response, "Custom 404", status_code=404)

        response = mock_custom_404_error_page(self.misago_request, Http404())
        self.assertNotContains(response, "Custom 404", status_code=404)
        response = mock_custom_404_error_page(self.site_request, Http404())
        self.assertContains(response, "Custom 404", status_code=404)



from pathlib import Path

import pytest

from ...conf import settings
from ...core.utils import slugify
from ..avatars import dynamic
from ..datadownloads import request_user_data_download
from ..models import Avatar, DataDownload, User
from ..utils import hash_email


def test_username_and_slug_is_anonymized(user):
    user.anonymize_data(anonymous_username="Deleted")
    assert user.username == "Deleted"
    assert user.slug == slugify("Deleted")


def test_user_avatar_files_are_deleted_during_user_deletion(user):
    dynamic.set_avatar(user)
    user.save()

    user_avatars = []
    for avatar in user.avatar_set.all():
        avatar_path = Path(avatar.image.path)
        assert avatar_path.exists()
        assert avatar_path.is_file()
        user_avatars.append(avatar)
    assert user_avatars

    user.delete(anonymous_username="Deleted")

    for removed_avatar in user_avatars:
        avatar_path = Path(removed_avatar.image.path)
        assert not avatar_path.exists()
        assert not avatar_path.is_file()

        with pytest.raises(Avatar.DoesNotExist):
            Avatar.objects.get(pk=removed_avatar.pk)


def test_username_setter_also_sets_slug():
    user = User()
    user.set_username("TestUser")
    assert user.username == "TestUser"
    assert user.slug == "testuser"


def test_django_username_getters_return_username(user):
    assert user.get_username() == user.username
    assert user.get_full_name() == user.username
    assert user.get_short_name() == user.username


def test_email_setter_normalizes_email():
    user = User()
    user.set_email("us3R@EXample.com")
    assert user.email == "us3R@example.com"


def test_email_setter_also_sets_email_hash():
    user = User()
    user.set_email("us3R@example.com")
    assert user.email_hash == hash_email("us3R@example.com")


def test_real_name_getter_returns_name_entered_in_profile_field(user):
    user.profile_fields["real_name"] = "John Doe"
    assert user.get_real_name() == "John Doe"


def test_real_name_getter_returns_none_if_profile_field_has_no_value(user):
    assert user.get_real_name() is None


def test_marking_user_for_deletion_deactivates_their_account_in_db(user):
    user.mark_for_delete()
    assert not user.is_active
    assert user.is_deleting_account

    user.refresh_from_db()
    assert not user.is_active
    assert user.is_deleting_account


def test_user_data_downloads_are_removed_by_anonymization(user):
    data_download = request_user_data_download(user)
    user.anonymize_data(anonymous_username="Deleted")

    with pytest.raises(DataDownload.DoesNotExist):
        data_download.refresh_from_db()


def test_deleting_user_also_deletes_their_data_downloads(user):
    data_download = request_user_data_download(user)
    user.delete(anonymous_username="Deleted")

    with pytest.raises(DataDownload.DoesNotExist):
        data_download.refresh_from_db()



# pylint: disable=line-too-long
"""
Default Misago settings. Override these with settings in the module pointed to
by the DJANGO_SETTINGS_MODULE environment variable.

If you rely on any of those in your code, make sure you use `misago.conf.settings`
instead of Django's `django.conf.settings`.
"""

# Permissions system extensions
# https://misago.readthedocs.io/en/latest/developers/acls.html#extending-permissions-system

MISAGO_ACL_EXTENSIONS = [
    "misago.users.permissions.account",
    "misago.users.permissions.profiles",
    "misago.users.permissions.moderation",
    "misago.users.permissions.delete",
    "misago.categories.permissions",
    "misago.threads.permissions.attachments",
    "misago.threads.permissions.polls",
    "misago.threads.permissions.threads",
    "misago.threads.permissions.privatethreads",
    "misago.threads.permissions.bestanswers",
    "misago.search.permissions",
]


# Path to the directory that Misago should use to prepare user data downloads.
# Should not be accessible from internet.

MISAGO_USER_DATA_DOWNLOADS_WORKING_DIR = None


# Custom markup extensions

MISAGO_MARKUP_EXTENSIONS = []


# Custom post validators

MISAGO_POST_VALIDATORS = []


# Post search filters

MISAGO_POST_SEARCH_FILTERS = []


# Posting middlewares
# https://misago.readthedocs.io/en/latest/developers/posting_process.html

MISAGO_POSTING_MIDDLEWARES = [
    # Always keep FloodProtectionMiddleware middleware first one
    "misago.threads.api.postingendpoint.floodprotection.FloodProtectionMiddleware",
    "misago.threads.api.postingendpoint.category.CategoryMiddleware",
    "misago.threads.api.postingendpoint.privatethread.PrivateThreadMiddleware",
    "misago.threads.api.postingendpoint.reply.ReplyMiddleware",
    "misago.threads.api.postingendpoint.moderationqueue.ModerationQueueMiddleware",
    "misago.threads.api.postingendpoint.attachments.AttachmentsMiddleware",
    "misago.threads.api.postingendpoint.participants.ParticipantsMiddleware",
    "misago.threads.api.postingendpoint.pin.PinMiddleware",
    "misago.threads.api.postingendpoint.close.CloseMiddleware",
    "misago.threads.api.postingendpoint.hide.HideMiddleware",
    "misago.threads.api.postingendpoint.protect.ProtectMiddleware",
    "misago.threads.api.postingendpoint.recordedit.RecordEditMiddleware",
    "misago.threads.api.postingendpoint.updatestats.UpdateStatsMiddleware",
    "misago.threads.api.postingendpoint.mentions.MentionsMiddleware",
    "misago.threads.api.postingendpoint.syncprivatethreads.SyncPrivateThreadsMiddleware",
    # Always keep SaveChangesMiddleware middleware after all state-changing middlewares
    "misago.threads.api.postingendpoint.savechanges.SaveChangesMiddleware",
    # Those middlewares are last because they don't change app state
    "misago.threads.api.postingendpoint.notifications.NotificationsMiddleware",
]


# Configured thread types

MISAGO_THREAD_TYPES = [
    "misago.threads.threadtypes.thread.Thread",
    "misago.threads.threadtypes.privatethread.PrivateThread",
]


# Search extensions

MISAGO_SEARCH_EXTENSIONS = [
    "misago.threads.search.SearchThreads",
    "misago.users.search.SearchUsers",
]


# Additional registration validators
# https://misago.readthedocs.io/en/latest/developers/validating_registrations.html

MISAGO_NEW_REGISTRATIONS_VALIDATORS = [
    "misago.users.validators.validate_gmail_email",
    "misago.users.validators.validate_with_sfs",
]


# Custom profile fields

MISAGO_PROFILE_FIELDS = []


# Login API URL

MISAGO_LOGIN_API_URL = "auth"


# Misago Admin Path
# Omit starting and trailing slashes. To disable Misago admin, empty this value.

MISAGO_ADMIN_PATH = "admincp"


# Admin urls namespaces that Misago's AdminAuthMiddleware should protect

MISAGO_ADMIN_NAMESPACES = ["admin", "misago:admin"]


# How long (in minutes) since previous request to admin namespace should admin session last.

MISAGO_ADMIN_SESSION_EXPIRATION = 60


# Display threads on forum index
# Change this to false to display categories list instead

MISAGO_THREADS_ON_INDEX = True


# How many notifications may be retrieved from the API in single request?

MISAGO_NOTIFICATIONS_PAGE_LIMIT = 50


# How many unread notifications to track
# Misago will not report report unread notifications count bigger than this
# Example: if limit 50 and user has 56 unread notifications, UI will show "50+"
# Also used by the notifications healing mechanism

MISAGO_UNREAD_NOTIFICATIONS_LIMIT = 50


# Function used for generating individual avatar for user

MISAGO_DYNAMIC_AVATAR_DRAWER = "misago.users.avatars.dynamic.draw_default"


# Path to directory containing avatar galleries
# Those galleries can be loaded by running loadavatargallery command

MISAGO_AVATAR_GALLERY = None


# Save user avatars for sizes
# Keep sizes ordered from greatest to smallest
# Max size also controls min size of uploaded image as well as crop size

MISAGO_AVATARS_SIZES = [400, 200, 150, 128, 100, 64, 50, 40, 32, 20]


# Path to blank avatar image used for guests and removed users.

MISAGO_BLANK_AVATAR = "misago/img/blank-avatar.png"


# Max allowed size of image before Misago will generate thumbnail for it

MISAGO_ATTACHMENT_IMAGE_SIZE_LIMIT = (500, 500)


# Length of secret used for attachments url tokens and filenames

MISAGO_ATTACHMENT_SECRET_LENGTH = 64


# Names of files served when user requests file that doesn't exist or is unavailable

MISAGO_ATTACHMENT_403_IMAGE = "misago/img/attachment-403.png"
MISAGO_ATTACHMENT_404_IMAGE = "misago/img/attachment-404.png"


# Available Moment.js locales

MISAGO_MOMENT_JS_LOCALES = [
    "af",
    "ar-ma",
    "ar-sa",
    "ar-tn",
    "ar",
    "az",
    "be",
    "bg",
    "bn",
    "bo",
    "br",
    "bs",
    "ca",
    "cs",
    "cv",
    "cy",
    "da",
    "de-at",
    "de",
    "el",
    "en-au",
    "en-ca",
    "en-gb",
    "eo",
    "es",
    "et",
    "eu",
    "fa",
    "fi",
    "fo",
    "fr-ca",
    "fr",
    "fy",
    "gl",
    "he",
    "hi",
    "hr",
    "hu",
    "hy-am",
    "id",
    "is",
    "it",
    "ja",
    "ka",
    "km",
    "ko",
    "lb",
    "lt",
    "lv",
    "mk",
    "ml",
    "mr",
    "ms-my",
    "my",
    "nb",
    "ne",
    "nl",
    "nn",
    "pl",
    "pt-br",
    "pt",
    "ro",
    "ru",
    "sk",
    "sl",
    "sq",
    "sr-cyrl",
    "sr",
    "sv",
    "ta",
    "th",
    "tl-ph",
    "tr",
    "tzm-latn",
    "tzm",
    "uk",
    "uz",
    "vi",
    "zh-cn",
    "zh-hans",
    "zh-tw",
]



from ..markup import checksums, signature_flavour


def set_user_signature(request, user, user_acl, signature):
    user.signature = signature

    if signature:
        user.signature_parsed = signature_flavour(request, user, user_acl, signature)
        user.signature_checksum = make_signature_checksum(user.signature_parsed, user)
    else:
        user.signature_parsed = ""
        user.signature_checksum = ""


def is_user_signature_valid(user):
    if user.signature:
        valid_checksum = make_signature_checksum(user.signature_parsed, user)
        return user.signature_checksum == valid_checksum
    return False


def make_signature_checksum(parsed_signature, user):
    return checksums.make_checksum(parsed_signature, [user.pk])



import pytest
from django.contrib.auth import get_user_model

User = get_user_model()


def test_get_user_by_username(db):
    user = User.objects.create_user("User", "test@example.com")
    assert User.objects.get_by_username("User") == user


def test_getting_user_by_username_is_case_insensitive(db):
    user = User.objects.create_user("User", "test@example.com")
    assert User.objects.get_by_username("uSeR") == user


def test_getting_user_by_username_raises_does_not_exist_for_no_result(db):
    with pytest.raises(User.DoesNotExist):
        User.objects.get_by_username("user")


def test_getting_user_by_username_supports_diacritics(db):
    with pytest.raises(User.DoesNotExist):
        User.objects.get_by_username("łóć")


def test_getting_user_by_username_is_not_doing_fuzzy_matching(db):
    User.objects.create_user("User", "test@example.com")
    with pytest.raises(User.DoesNotExist):
        User.objects.get_by_username("usar")


def test_get_user_by_email(db):
    user = User.objects.create_user("User", "test@example.com")
    assert User.objects.get_by_email("test@example.com") == user


def test_getting_user_by_email_is_case_insensitive(db):
    user = User.objects.create_user("User", "test@example.com")
    assert User.objects.get_by_email("tEsT@eXaMplE.com") == user


def test_getting_user_by_email_supports_diacritics(db):
    user = User.objects.create_user("User", "łóć@łexĄmple.com")
    assert User.objects.get_by_email("łÓć@ŁexĄMple.com") == user


def test_getting_user_by_email_raises_does_not_exist_for_no_result(db):
    with pytest.raises(User.DoesNotExist):
        User.objects.get_by_email("test@example.com")


def test_get_user_by_username_using_combined_getter(db):
    user = User.objects.create_user("User", "test@example.com")
    assert User.objects.get_by_username_or_email("user") == user


def test_get_user_by_email_using_combined_getter(db):
    user = User.objects.create_user("User", "test@example.com")
    assert User.objects.get_by_username_or_email("test@example.com") == user


def test_combined_getter_handles_username_slug_and_email_collision(db):
    email_match = User.objects.create_user("Bob", "test@test.test")
    slug_match = User.objects.create_user("TestTestTest", "other@test.com")

    assert User.objects.get_by_username_or_email("test@test.test") == email_match
    assert User.objects.get_by_username_or_email("TestTestTest") == slug_match


def test_combined_getter_raises_does_not_exist_for_no_result(db):
    with pytest.raises(User.DoesNotExist):
        User.objects.get_by_username_or_email("User")



from ..client import get_value_from_json


def test_json_str_value_is_returned():
    assert get_value_from_json("val", {"val": "ok", "val2": "nope"}) == "ok"


def test_json_str_value_has_stripped_whitespace():
    assert get_value_from_json("val", {"val": "  ok  ", "val2": "nope"}) == "ok"


def test_json_int_value_is_cast_to_str():
    assert get_value_from_json("val", {"val": 21, "val2": "nope"}) == "21"


def test_none_is_returned_if_val_is_not_found():
    assert get_value_from_json("val", {"val3": 21, "val2": "nope"}) is None


def test_none_is_returned_if_val_is_null():
    assert get_value_from_json("val", {"val": None, "val2": "nope"}) is None


def test_json_value_is_returned_from_nested_object():
    assert (
        get_value_from_json(
            "val.child.val",
            {
                "val2": "nope",
                "val": {
                    "child": {
                        "val2": "nope",
                        "val": "ok",
                    },
                },
            },
        )
        == "ok"
    )


def test_none_is_returned_from_nested_object_missing_value():
    assert (
        get_value_from_json(
            "val.child.val3",
            {
                "val2": "nope",
                "val": {
                    "child": {
                        "val2": "nope",
                        "val": "ok",
                    },
                },
            },
        )
        is None
    )


def test_json_value_returned_from_nested_object_is_str():
    assert (
        get_value_from_json(
            "val.child.val",
            {
                "val2": "nope",
                "val": {
                    "child": {
                        "val2": "nope",
                        "val": 21,
                    },
                },
            },
        )
        == "21"
    )


def test_none_is_returned_from_nested_object_missing_path_item():
    assert (
        get_value_from_json(
            "val.missing.val",
            {
                "val2": "nope",
                "val": {
                    "child": {
                        "val2": "nope",
                        "val": 21,
                    },
                },
            },
        )
        is None
    )


def test_none_is_returned_from_nested_path_item_not_being_dict():
    assert (
        get_value_from_json(
            "val.child.val",
            {
                "val2": "nope",
                "val": {
                    "child": "nope",
                },
            },
        )
        is None
    )



from django.utils.translation import pgettext_lazy


class OAuth2Error(Exception):
    pass


class OAuth2ProviderError(OAuth2Error):
    pass


class OAuth2AccessDeniedError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error", "The OAuth2 process was canceled by the provider."
    )


class OAuth2StateError(OAuth2Error):
    recoverable = True


class OAuth2StateNotSetError(OAuth2StateError):
    message = pgettext_lazy("oauth2 error", "The OAuth2 session is missing state.")


class OAuth2StateNotProvidedError(OAuth2StateError):
    message = pgettext_lazy(
        "oauth2 error", "The OAuth2 state was not sent by the provider."
    )


class OAuth2StateMismatchError(OAuth2StateError):
    message = pgettext_lazy(
        "oauth2 error",
        "The OAuth2 state sent by the provider did not match one in the session.",
    )


class OAuth2CodeError(OAuth2Error):
    recoverable = True


class OAuth2CodeNotProvidedError(OAuth2CodeError):
    message = pgettext_lazy(
        "oauth2 error", "The OAuth2 authorization code was not sent by the provider."
    )


class OAuth2ProviderError(OAuth2Error):
    recoverable = True


class OAuth2AccessTokenRequestError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "Failed to connect to the OAuth2 provider to retrieve an access token.",
    )


class OAuth2AccessTokenResponseError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "The OAuth2 provider responded with error for an access token request.",
    )


class OAuth2AccessTokenJSONError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "The OAuth2 provider did not respond with a valid JSON for an access token request.",
    )


class OAuth2AccessTokenNotProvidedError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "JSON sent by the OAuth2 provider did not contain an access token.",
    )


class OAuth2UserDataRequestError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "Failed to connect to the OAuth2 provider to retrieve user profile.",
    )


class OAuth2UserDataResponseError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "The OAuth2 provider responded with error for user profile request.",
    )


class OAuth2UserDataJSONError(OAuth2ProviderError):
    message = pgettext_lazy(
        "oauth2 error",
        "The OAuth2 provider did not respond with a valid JSON for user profile request.",
    )


class OAuth2UserIdNotProvidedError(OAuth2Error):
    message = pgettext_lazy(
        "oauth2 error", "JSON sent by the OAuth2 provider did not contain a user ID."
    )


class OAuth2UserAccountDeactivatedError(OAuth2Error):
    recoverable = False
    message = pgettext_lazy(
        "oauth2 error",
        "User account associated with the profile from the OAuth2 provider was deactivated by the site administrator.",
    )


class OAuth2UserDataValidationError(OAuth2ProviderError):
    recoverable = False
    error_list: list[str]
    message = pgettext_lazy(
        "oauth2 error",
        "User profile retrieved from the OAuth2 provider did not validate.",
    )

    def __init__(self, error_list: list[str]):
        self.error_list = error_list



from ..checksums import is_checksum_valid, make_checksum

message = "Test message."
post_pk = 123


def test_checksum_can_be_generated_for_post_message_and_pk():
    assert make_checksum(message, [post_pk])


def test_valid_message_checksum_is_checked():
    checksum = make_checksum(message, [post_pk])
    assert is_checksum_valid(message, checksum, [post_pk])


def test_checksum_invalidates_if_message_is_changed():
    checksum = make_checksum(message, [post_pk])
    assert not is_checksum_valid("Changed message.", checksum, [post_pk])


def test_checksum_invalidates_if_pk_is_changed():
    checksum = make_checksum(message, [post_pk])
    assert not is_checksum_valid(message, checksum, [post_pk + 1])



from django.urls import reverse

from ....acl.models import Role
from ....admin.test import AdminTestCase
from ...models import AttachmentType


class AttachmentTypeAdminViewsTests(AdminTestCase):
    def setUp(self):
        super().setUp()
        self.admin_link = reverse("misago:admin:settings:attachment-types:index")

    def test_link_registered(self):
        """admin nav contains attachment types link"""
        response = self.client.get(reverse("misago:admin:settings:index"))
        self.assertContains(response, self.admin_link)

    def test_list_view(self):
        """attachment types list returns 200 and renders all attachment types"""
        response = self.client.get(self.admin_link)
        self.assertEqual(response.status_code, 200)

        for attachment_type in AttachmentType.objects.all():
            self.assertContains(response, attachment_type.name)
            for file_extension in attachment_type.extensions_list:
                self.assertContains(response, file_extension)
            for mimename in attachment_type.mimetypes_list:
                self.assertContains(response, mimename)

    def test_new_view(self):
        """new attachment type view has no showstoppers"""
        form_link = reverse("misago:admin:settings:attachment-types:new")

        response = self.client.get(form_link)
        self.assertEqual(response.status_code, 200)

        response = self.client.post(form_link, data={})
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            form_link,
            data={
                "name": "Test type",
                "extensions": ".test",
                "size_limit": 0,
                "status": AttachmentType.ENABLED,
            },
        )
        self.assertEqual(response.status_code, 302)

        # clean alert about new item created
        self.client.get(self.admin_link)

        response = self.client.get(self.admin_link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Test type")
        self.assertContains(response, "test")

    def test_edit_view(self):
        """edit attachment type view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:settings:attachment-types:new"),
            data={
                "name": "Test type",
                "extensions": ".test",
                "size_limit": 0,
                "status": AttachmentType.ENABLED,
            },
        )

        test_type = AttachmentType.objects.order_by("id").last()
        self.assertEqual(test_type.name, "Test type")

        form_link = reverse(
            "misago:admin:settings:attachment-types:edit", kwargs={"pk": test_type.pk}
        )

        response = self.client.get(form_link)
        self.assertEqual(response.status_code, 200)

        response = self.client.post(form_link, data={})
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            form_link,
            data={
                "name": "Test type edited",
                "extensions": ".test.extension",
                "mimetypes": "test/edited-mime",
                "size_limit": 512,
                "status": AttachmentType.DISABLED,
                "limit_uploads_to": [r.pk for r in Role.objects.all()],
                "limit_downloads_to": [r.pk for r in Role.objects.all()],
            },
        )
        self.assertEqual(response.status_code, 302)

        test_type = AttachmentType.objects.order_by("id").last()
        self.assertEqual(test_type.name, "Test type edited")

        # clean alert about new item created
        self.client.get(self.admin_link)

        response = self.client.get(self.admin_link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, test_type.name)
        self.assertContains(response, test_type.extensions)
        self.assertContains(response, test_type.mimetypes)

        self.assertEqual(test_type.limit_uploads_to.count(), Role.objects.count())
        self.assertEqual(test_type.limit_downloads_to.count(), Role.objects.count())

        # remove limits from type
        response = self.client.post(
            form_link,
            data={
                "name": "Test type edited",
                "extensions": ".test.extension",
                "mimetypes": "test/edited-mime",
                "size_limit": 512,
                "status": AttachmentType.DISABLED,
                "limit_uploads_to": [],
                "limit_downloads_to": [],
            },
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(test_type.limit_uploads_to.count(), 0)
        self.assertEqual(test_type.limit_downloads_to.count(), 0)

    def test_clean_params_view(self):
        """admin form nicely cleans lists of extensions/mimetypes"""
        TEST_CASES = [
            ("test", ["test"]),
            (".test", ["test"]),
            (".tar.gz", ["tar.gz"]),
            (". test", ["test"]),
            ("test, test", ["test"]),
            ("test, tEst", ["test"]),
            ("test, other, tEst", ["test", "other"]),
            ("test, other, tEst,OTher", ["test", "other"]),
        ]

        for raw, final in TEST_CASES:
            response = self.client.post(
                reverse("misago:admin:settings:attachment-types:new"),
                data={
                    "name": "Test type",
                    "extensions": raw,
                    "size_limit": 0,
                    "status": AttachmentType.ENABLED,
                },
            )
            self.assertEqual(response.status_code, 302)

            test_type = AttachmentType.objects.order_by("id").last()
            self.assertEqual(set(test_type.extensions_list), set(final))

    def test_delete_view(self):
        """delete attachment type view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:settings:attachment-types:new"),
            data={
                "name": "Test type",
                "extensions": ".test",
                "size_limit": 0,
                "status": AttachmentType.ENABLED,
            },
        )

        test_type = AttachmentType.objects.order_by("id").last()
        self.assertEqual(test_type.name, "Test type")

        action_link = reverse(
            "misago:admin:settings:attachment-types:delete", kwargs={"pk": test_type.pk}
        )

        response = self.client.post(action_link)
        self.assertEqual(response.status_code, 302)

        # clean alert about item deleted
        self.client.get(self.admin_link)

        response = self.client.get(self.admin_link)
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, test_type.name)

    def test_cant_delete_type_with_attachments_view(self):
        """delete attachment type is not allowed if it has attachments associated"""
        self.client.post(
            reverse("misago:admin:settings:attachment-types:new"),
            data={
                "name": "Test type",
                "extensions": ".test",
                "size_limit": 0,
                "status": AttachmentType.ENABLED,
            },
        )

        test_type = AttachmentType.objects.order_by("id").last()
        self.assertEqual(test_type.name, "Test type")

        test_type.attachment_set.create(
            secret="loremipsum",
            filetype=test_type,
            uploader_name="User",
            uploader_slug="user",
            filename="test.zip",
            file="sad76asd678as687sa.zip",
        )

        action_link = reverse(
            "misago:admin:settings:attachment-types:delete", kwargs={"pk": test_type.pk}
        )

        response = self.client.post(action_link)
        self.assertEqual(response.status_code, 302)

        # get alert form database
        AttachmentType.objects.get(pk=test_type.pk)



from datetime import timedelta
from io import StringIO

from django.core import management
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..management.commands import clearnotifications
from ..models import Notification


def call_command():
    command = clearnotifications.Command()

    out = StringIO()
    management.call_command(command, stdout=out)
    return out.getvalue().strip().splitlines()[-1].strip()


def test_command_works_if_there_are_no_notifications(db):
    command_output = call_command()
    assert command_output == "No old notifications have been deleted."


@override_dynamic_settings(delete_notifications_older_than=5)
def test_recent_notification_is_kept(user, post):
    Notification.objects.create(user=user, verb="TEST")

    command_output = call_command()
    assert command_output == "No old notifications have been deleted."
    assert Notification.objects.exists()


@override_dynamic_settings(delete_notifications_older_than=5)
def test_old_notification_is_deleted(user, post):
    Notification.objects.create(user=user, verb="TEST")
    Notification.objects.update(created_at=timezone.now() - timedelta(days=10))

    command_output = call_command()
    assert command_output == "Deleted 1 old notifications."
    assert not Notification.objects.exists()



import os

import pytest
from django.urls import reverse

from ....test import assert_contains
from ...models import Theme

import_link = reverse("misago:admin:themes:import")


class MockThemeExport:
    def __init__(self, response):
        self.name = "theme-export.zip"
        self.content_type = response["content-type"]
        self.size = response["content-length"]
        self._response = response

    def read(self):
        return self._response.getvalue()


@pytest.fixture
def reimport_theme(admin_client):
    def export_import_theme(theme, extra_data=None):
        export_link = reverse("misago:admin:themes:export", kwargs={"pk": theme.pk})
        theme_export = MockThemeExport(admin_client.post(export_link))

        data = extra_data or {}
        data["upload"] = theme_export
        admin_client.post(import_link, data)

        return Theme.objects.filter(pk__gt=theme.pk).last()

    return export_import_theme


def assert_filenames_are_same(src, dst):
    source_filename = os.path.split(src.path)[-1]
    imported_filename = os.path.split(dst.path)[-1]
    assert source_filename == imported_filename


def test_import_theme_form_is_displayed(admin_client):
    response = admin_client.get(import_link)
    assert_contains(response, "Import theme")


def test_theme_import_fails_if_export_was_not_uploaded(admin_client):
    admin_client.post(import_link, {"upload": ""})
    assert Theme.objects.count() == 1


def test_empty_theme_export_can_be_imported_back(reimport_theme, theme):
    assert reimport_theme(theme)


def test_theme_can_be_imported_with_custom_name(reimport_theme, theme):
    imported_theme = reimport_theme(theme, {"name": "Imported theme"})
    assert imported_theme.name == "Imported theme"


def test_theme_can_be_imported_without_parent(reimport_theme, theme):
    imported_theme = reimport_theme(theme)
    assert imported_theme.parent is None


def test_theme_can_be_imported_with_custom_parent(reimport_theme, theme):
    imported_theme = reimport_theme(theme, {"parent": theme.pk})
    assert imported_theme.parent == theme


def test_theme_can_be_imported_with_default_theme_asparent(
    reimport_theme, theme, default_theme
):
    imported_theme = reimport_theme(theme, {"parent": default_theme.pk})
    assert imported_theme.parent == default_theme


def test_theme_import_fails_if_parent_is_nonexisisting(
    reimport_theme, theme, nonexisting_theme
):
    assert not reimport_theme(theme, {"parent": nonexisting_theme.pk})


def test_importing_theme_under_parent_rebuilds_themes_tree(reimport_theme, theme):
    imported_theme = reimport_theme(theme, {"parent": theme.pk})
    theme.refresh_from_db()

    assert imported_theme.tree_id == theme.tree_id
    assert theme.lft == 1
    assert theme.rght == 4
    assert imported_theme.lft == 2
    assert imported_theme.rght == 3


def test_theme_details_are_exported_and_imported_back(reimport_theme, theme):
    theme.name = "Exported Theme"
    theme.version = "0.1.2 FINAL"
    theme.author = "John Doe"
    theme.url = "https://example.com"
    theme.save()

    imported_theme = reimport_theme(theme)

    assert imported_theme.name == theme.name
    assert imported_theme.version == theme.version
    assert imported_theme.author == theme.author
    assert imported_theme.url == theme.url


def test_imported_theme_has_own_dirname(reimport_theme, theme):
    imported_theme = reimport_theme(theme)
    assert theme.dirname != imported_theme.dirname


def test_css_file_is_exported_and_imported_back(reimport_theme, theme, css):
    imported_theme = reimport_theme(theme)

    imported_css = imported_theme.css.last()
    assert imported_css.name == css.name
    assert imported_css.source_hash == css.source_hash
    assert imported_css.size == css.size

    assert_filenames_are_same(css.source_file, imported_css.source_file)


def test_css_link_is_exported_and_imported_back(reimport_theme, theme, css_link):
    imported_theme = reimport_theme(theme)

    imported_css = imported_theme.css.last()
    assert imported_css.name == css_link.name
    assert imported_css.url == css_link.url


def test_importing_css_link_triggers_getting_remote_css_size_task(
    reimport_theme, theme, css_link, mock_update_remote_css_size
):
    reimport_theme(theme)
    mock_update_remote_css_size.assert_called_once()


def test_theme_export_containing_css_file_and_link_can_be_imported_back(
    reimport_theme, theme, css, css_link
):
    imported_theme = reimport_theme(theme)
    css_names = list(imported_theme.css.values_list("name", flat=True))
    assert css_names == [css.name, css_link.name]


def test_theme_css_order_is_preserved_after_import(
    reimport_theme, theme, css, css_link
):
    css_link.order = 1
    css_link.save()

    css.order = 2
    css.save()

    imported_theme = reimport_theme(theme)
    css_names = list(imported_theme.css.values_list("name", flat=True))
    assert css_names == [css_link.name, css.name]


def test_theme_export_containing_media_file_can_be_imported_back(
    reimport_theme, theme, media
):
    imported_theme = reimport_theme(theme)

    imported_media = imported_theme.media.last()
    assert imported_media.name == media.name
    assert imported_media.hash == media.hash
    assert imported_media.type == media.type
    assert imported_media.size == media.size

    assert_filenames_are_same(media.file, imported_media.file)


def test_theme_export_containing_image_file_can_be_imported_back(
    reimport_theme, theme, image
):
    imported_theme = reimport_theme(theme)

    imported_image = imported_theme.media.last()
    assert imported_image.name == image.name
    assert imported_image.hash == image.hash
    assert imported_image.type == image.type
    assert imported_image.width == image.width
    assert imported_image.height == image.height
    assert imported_image.size == image.size

    assert_filenames_are_same(image.file, imported_image.file)


def test_theme_export_containing_different_files_can_be_imported_back(
    reimport_theme, theme, css, css_link, media, image
):
    imported_theme = reimport_theme(theme)

    css_names = list(imported_theme.css.values_list("name", flat=True))
    assert css_names == [css.name, css_link.name]

    media_names = list(imported_theme.media.values_list("name", flat=True))
    assert media_names == [image.name, media.name]


def test_importing_theme_triggers_css_build(
    reimport_theme, theme, css_link, mock_build_theme_css
):
    imported_theme = reimport_theme(theme)
    mock_build_theme_css.assert_called_once_with(imported_theme.pk)



from django.db import transaction
from rest_framework.response import Response

from ...moderation import threads as moderation
from ...permissions import allow_delete_thread
from ...serializers import DeleteThreadsSerializer


@transaction.atomic
def delete_thread(request, thread):
    allow_delete_thread(request.user_acl, thread)
    moderation.delete_thread(request, thread)
    return Response({})


def delete_bulk(request, viewmodel):
    serializer = DeleteThreadsSerializer(
        data={"threads": request.data},
        context={
            "request": request,
            "settings": request.settings,
            "viewmodel": viewmodel,
        },
    )

    if not serializer.is_valid():
        if "threads" in serializer.errors:
            errors = serializer.errors["threads"]
            if "details" in errors:
                return Response(hydrate_error_details(errors["details"]), status=400)
            # Fix for KeyError - errors[0]
            try:
                return Response({"detail": errors[0]}, status=403)
            except KeyError:
                return Response({"detail": list(errors.values())[0][0]}, status=403)

        errors = list(serializer.errors)[0][0]
        return Response({"detail": errors}, status=400)

    for thread in serializer.validated_data["threads"]:
        with transaction.atomic():
            delete_thread(request, thread)

    return Response([])


def hydrate_error_details(errors):
    for error in errors:
        error["thread"]["id"] = int(error["thread"]["id"])
    return errors



from rest_framework import serializers

from . import PostingEndpoint, PostingMiddleware
from ... import moderation


class CloseMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return self.mode == PostingEndpoint.START

    def get_serializer(self):
        return CloseSerializer(data=self.request.data)

    def post_save(self, serializer):
        if self.thread.category.acl["can_close_threads"]:
            if serializer.validated_data.get("close"):
                moderation.close_thread(self.request, self.thread)


class CloseSerializer(serializers.Serializer):
    close = serializers.BooleanField(required=False, default=False)



from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext
from rest_framework.response import Response

from ....acl.objectacl import add_acl_to_obj
from ...serializers import MergePostsSerializer, PostSerializer


def posts_merge_endpoint(request, thread):
    if not thread.acl["can_merge_posts"]:
        raise PermissionDenied(
            pgettext("posts api", "You can't merge posts in this thread.")
        )

    serializer = MergePostsSerializer(
        data=request.data,
        context={
            "settings": request.settings,
            "thread": thread,
            "user_acl": request.user_acl,
        },
    )

    if not serializer.is_valid():
        # Fix for KeyError - errors[0]
        errors = list(serializer.errors.values())[0]
        try:
            return Response({"detail": errors[0]}, status=400)
        except KeyError:
            return Response({"detail": list(errors.values())[0][0]}, status=400)

    posts = serializer.validated_data["posts"]
    first_post, merged_posts = posts[0], posts[1:]

    for post in merged_posts:
        post.merge(first_post)
        post.delete()

    if first_post.pk == thread.first_post_id:
        first_post.set_search_document(thread.title)
    else:
        first_post.set_search_document()

    first_post.save()

    first_post.update_search_vector()
    first_post.save(update_fields=["search_vector"])

    first_post.postread_set.all().delete()

    thread.synchronize()
    thread.save()

    thread.category.synchronize()
    thread.category.save()

    first_post.thread = thread
    first_post.category = thread.category

    add_acl_to_obj(request.user_acl, first_post)

    return Response(PostSerializer(first_post, context={"user": request.user}).data)



from io import StringIO

from django.core.management import call_command

from ...notifications.models import Notification
from ..management.commands import createfakenotifications


def test_management_command_creates_notifications(
    user, other_user, thread, reply, other_thread
):
    call_command(createfakenotifications.Command(), stdout=StringIO())
    assert Notification.objects.count() == 6  # 2 users x 3 posts



import html
from typing import TYPE_CHECKING, Callable, Dict, overload

from django.http import HttpRequest
from django.utils.translation import pgettext

from ..categories.enums import CategoryTree
from ..threads.views.goto import PrivateThreadGotoPostView, ThreadGotoPostView
from .verbs import NotificationVerb
from .exceptions import NotificationVerbError

if TYPE_CHECKING:
    from .models import Notification


MessageFactory = Callable[["Notification"], str]
RedirectFactory = Callable[["HttpRequest", "Notification"], str]


class NotificationRegistry:
    _messages: Dict[str, MessageFactory]
    _redirects: Dict[str, RedirectFactory]

    def __init__(self):
        self._messages = {}
        self._redirects = {}

    @overload
    def message(self, verb: str) -> Callable[[MessageFactory], MessageFactory]:
        ...

    def message(self, verb: str, factory: MessageFactory | None = None) -> None:
        """Register `factory` function as message factory.

        Can be called with two arguments or used as decorator:

        ```python
        registry.message("replied", get_replied_message)

        # ...or...

        @registry.message("replied")
        def get_replied_message(notification: Notification) -> str:
            ...
        ```
        """
        if factory:
            self._messages[verb] = factory
            return

        def register_factory(f: MessageFactory):
            self._messages[verb] = f
            return f

        return register_factory

    def get_message(self, notification: "Notification") -> str:
        """Returns `str` with message for given notification.

        Raises `NotificationVerbError` when there's no message function for
        given notification type.

        ```python
        message = registry.get_message(notification)
        ```
        """
        try:
            return self._messages[notification.verb](notification)
        except KeyError as exc:
            raise NotificationVerbError(notification.verb) from exc

    @overload
    def redirect(self, verb: str) -> Callable[[RedirectFactory], RedirectFactory]:
        ...

    def redirect(self, verb: str, factory: RedirectFactory | None = None) -> None:
        """Register `factory` function as redirect url factory.

        Can be called with two arguments or used as decorator:

        ```python
        registry.redirect("replied", get_replied_post_url)

        # ...or...

        @registry.redirect("replied")
        def get_replied_post_url(notification: Notification) -> str:
            ...
        ```
        """
        if factory:
            self._redirects[verb] = factory
            return

        def register_factory(f: RedirectFactory):
            self._redirects[verb] = f
            return f

        return register_factory

    def get_redirect_url(
        self, request: HttpRequest, notification: "Notification"
    ) -> str:
        """Returns `str` with redirect url to given notification's target.

        Raises `NotificationVerbError` when there's no redirect function for
        given notification type.

        ```python
        message = registry.get_redirect_url(request, notification)
        ```
        """
        try:
            return self._redirects[notification.verb](request, notification)
        except KeyError as exc:
            raise NotificationVerbError(notification.verb) from exc


registry = NotificationRegistry()


# TEST: used in tests only


@registry.message("TEST")
def get_test_notification_message(notification: "Notification") -> str:
    return f"Test notification #{notification.id}"


@registry.redirect("TEST")
def get_test_notification_url(
    request: HttpRequest, notification: "Notification"
) -> str:
    return f"/#test-notification-{notification.id}"


# REPLIED: new reply in thread or private thread


@registry.message(NotificationVerb.REPLIED)
def get_replied_notification_message(notification: "Notification") -> str:
    message = html.escape(
        pgettext("notification replied", "%(actor)s replied to %(thread)s")
    )
    return message % {
        "actor": bold_escape(notification.actor_name),
        "thread": bold_escape(notification.thread_title),
    }


go_to_thread_post = ThreadGotoPostView.as_view()
go_to_private_thread_post = PrivateThreadGotoPostView.as_view()


@registry.redirect(NotificationVerb.REPLIED)
def get_replied_notification_url(
    request: HttpRequest, notification: "Notification"
) -> str:
    if notification.category.tree_id == CategoryTree.PRIVATE_THREADS:
        view = go_to_private_thread_post
    else:
        view = go_to_thread_post

    redirect = view(
        request,
        notification.thread_id,
        notification.thread.slug,
        post=notification.post_id,
    )

    return redirect.headers["location"]


# INVITED: invited to private thread


@registry.message(NotificationVerb.INVITED)
def get_invited_notification_message(notification: "Notification") -> str:
    message = html.escape(
        pgettext("notification invited", "%(actor)s invited you to %(thread)s")
    )
    return message % {
        "actor": bold_escape(notification.actor_name),
        "thread": bold_escape(notification.thread_title),
    }


@registry.redirect(NotificationVerb.INVITED)
def get_invited_notification_url(
    request: HttpRequest, notification: "Notification"
) -> str:
    return notification.category.thread_type.get_thread_absolute_url(
        notification.thread
    )


def bold_escape(value: str) -> str:
    return f"<b>{html.escape(value)}</b>"



from dataclasses import dataclass
from typing import Any, List, Optional

from django.http import Http404
from rest_framework.exceptions import ValidationError


class PaginationError(ValidationError):
    pass


@dataclass
class PaginationResult:
    items: List[Any]
    has_next: bool
    has_previous: bool
    first_cursor: Optional[Any]
    last_cursor: Optional[Any]


def paginate_queryset(
    request,
    queryset,
    order_by: str,
    max_limit: int,
    raise_404: bool = False,
) -> PaginationResult:
    after = get_query_value(request, "after")
    before = get_query_value(request, "before")
    limit = get_query_value(request, "limit") or max_limit

    if after and before:
        raise PaginationError(
            {"after": "'after' and 'before' can't be used at same time"}
        )

    if limit > max_limit:
        raise PaginationError({"limit": f"can't be greater than '{max_limit}'"})

    queryset = queryset.order_by(order_by)

    if before:
        return paginate_with_before(queryset, before, order_by, limit, raise_404)

    return paginate_with_after(queryset, after or 0, order_by, limit, raise_404)


def get_query_value(request, name: str, default: int | None = None) -> int | None:
    value = request.GET[name] if name in request.GET else default
    if value is None:
        return value

    try:
        value = int(value)
        if value < 1:
            raise ValueError()
    except (TypeError, ValueError):
        raise PaginationError({name: "must be a positive integer"})

    return value


def paginate_with_after(
    queryset,
    after: int,
    order_by: str,
    limit: int,
    raise_404: bool,
) -> PaginationResult:
    order_desc = order_by[0] == "-"
    col_name = order_by.lstrip("-")

    if after:
        if order_desc:
            filter_kwarg = {f"{col_name}__lt": after}
        else:
            filter_kwarg = {f"{col_name}__gt": after}
        items_queryset = queryset.filter(**filter_kwarg)
    else:
        items_queryset = queryset

    items = list(items_queryset[: limit + 1])
    has_next = False
    has_previous = False

    if after and not items and raise_404:
        raise Http404()

    if len(items) > limit:
        has_next = True
        items = items[:limit]

    if items:
        cursor = getattr(items[0], col_name)
        if order_desc:
            filter_kwarg = {f"{col_name}__gt": cursor}
        else:
            filter_kwarg = {f"{col_name}__lt": cursor}
        has_previous = queryset.filter(**filter_kwarg).order_by(col_name).exists()

    return create_pagination_result(
        col_name,
        items,
        has_next,
        has_previous,
    )


def paginate_with_before(
    queryset,
    before: int,
    order_by: str,
    limit: int,
    raise_404: bool,
) -> PaginationResult:
    order_desc = order_by[0] == "-"
    col_name = order_by.lstrip("-")

    if order_desc:
        filter_kwarg = {f"{col_name}__gt": before}
    else:
        filter_kwarg = {f"{col_name}__lt": before}
    items_queryset = queryset.filter(**filter_kwarg).order_by(
        col_name if order_desc else f"-{col_name}"
    )

    items = list(reversed(items_queryset[: limit + 1]))
    has_next = False
    has_previous = False

    if not items and raise_404:
        raise Http404()

    if len(items) > limit:
        items = items[1:]
        has_previous = True

    if items:
        cursor = getattr(items[-1], col_name)
        if order_desc:
            filter_kwarg = {f"{col_name}__lt": cursor}
        else:
            filter_kwarg = {f"{col_name}__gt": cursor}
        has_next = queryset.filter(**filter_kwarg).order_by(col_name).exists()

    return create_pagination_result(
        col_name,
        items,
        has_next,
        has_previous,
    )


def create_pagination_result(
    col_name: str,
    items: List[Any],
    has_next: bool,
    has_previous: bool,
) -> PaginationResult:
    if items:
        first_cursor = getattr(items[0], col_name)
        last_cursor = getattr(items[-1], col_name)
    else:
        first_cursor = None
        last_cursor = None

    return PaginationResult(
        items=items,
        has_next=has_next,
        has_previous=has_previous,
        first_cursor=first_cursor,
        last_cursor=last_cursor,
    )



from django.db import IntegrityError
from django.utils.translation import pgettext
from rest_framework import status
from rest_framework.response import Response

from ...namechanges import get_username_options
from ...serializers import ChangeUsernameSerializer


def username_endpoint(request):
    if request.method == "POST":
        return change_username(request)

    options = get_username_options_from_request(request)
    return options_response(options)


def get_username_options_from_request(request):
    return get_username_options(request.settings, request.user, request.user_acl)


def options_response(options):
    if options["next_on"]:
        options["next_on"] = options["next_on"].isoformat()
    return Response(options)


def change_username(request):
    options = get_username_options_from_request(request)
    if not options["changes_left"]:
        return Response(
            {
                "detail": pgettext(
                    "change username api",
                    "You can't change your username now.",
                ),
                "options": options,
            },
            status=status.HTTP_400_BAD_REQUEST,
        )

    serializer = ChangeUsernameSerializer(
        data=request.data, context={"settings": request.settings, "user": request.user}
    )
    if not serializer.is_valid():
        return Response(
            {"detail": serializer.errors["non_field_errors"][0]},
            status=status.HTTP_400_BAD_REQUEST,
        )

    try:
        serializer.change_username(changed_by=request.user)
        updated_options = get_username_options_from_request(request)
        if updated_options["next_on"]:
            updated_options["next_on"] = updated_options["next_on"].isoformat()

        return Response(
            {
                "username": request.user.username,
                "slug": request.user.slug,
                "options": updated_options,
            }
        )
    except IntegrityError:
        return Response(
            {
                "detail": pgettext(
                    "change username api",
                    "Error changing username. Please try again.",
                )
            },
            status=status.HTTP_400_BAD_REQUEST,
        )


def moderate_username_endpoint(request, profile):
    if request.method == "POST":
        serializer = ChangeUsernameSerializer(
            data=request.data, context={"settings": request.settings, "user": profile}
        )

        if not serializer.is_valid():
            return Response(
                {"detail": serializer.errors["non_field_errors"][0]},
                status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            serializer.change_username(changed_by=request.user)
            return Response({"username": profile.username, "slug": profile.slug})
        except IntegrityError:
            return Response(
                {
                    "detail": pgettext(
                        "change username api",
                        "Error changing username. Please try again.",
                    )
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

    return Response(
        {
            "length_min": request.settings.username_length_min,
            "length_max": request.settings.username_length_max,
        }
    )
